"""
Advances in Financial Machine Learning, Marcos Lopez de Prado
Chapter 2: Financial Data Structures

This module contains the functions to help users create structured financial data from raw unstructured data,
in the form of time, tick, volume, and dollar bars.

These bars are used throughout the text book (Advances in Financial Machine Learning, By Marcos Lopez de Prado, 2018,
pg 25) to build the more interesting features for predicting financial time series data.

These financial data structures have better statistical properties when compared to those based on fixed time interval
sampling. A great paper to read more about this is titled: The Volume Clock: Insights into the high frequency paradigm,
Lopez de Prado, et al.

Many of the projects going forward will require Dollar and Volume bars.
"""

# Imports
from typing import Union, Iterable, Optional

import numpy as np
import pandas as pd

from famlafl.data_structures.base_bars import BaseBars


class StandardBars(BaseBars):
    """
    Contains all of the logic to construct the standard bars from chapter 2. This class shouldn't be used directly.
    We have added functions to the package such as get_dollar_bars which will create an instance of this
    class and then construct the standard bars, to return to the user.

    This is because we wanted to simplify the logic as much as possible, for the end user.
    """

    def __init__(self, metric: str, threshold: int = 50000, batch_size: int = 20000000):
        """
        Constructor

        :param metric: (str) Type of run bar to create. Example: "dollar_run"
        :param threshold: (int) Threshold at which to sample
        :param batch_size: (int) Number of rows to read in from the csv, per batch
        """
        BaseBars.__init__(self, metric, batch_size)

        # Threshold at which to sample
        self.threshold = threshold

    def _reset_cache(self):
        """
        Implementation of abstract method _reset_cache for standard bars
        """
        self.open_price = None
        self.high_price, self.low_price = -np.inf, np.inf
        self.cum_statistics = {'cum_ticks': 0, 'cum_dollar_value': 0, 'cum_volume': 0, 'cum_buy_volume': 0}

    def _extract_bars(self, data: Union[list, tuple, np.ndarray]) -> list:
        """
        For loop which compiles the various bars: dollar, volume, or tick.
        We did investigate the use of trying to solve this in a vectorised manner but found that a For loop worked well.

        :param data: (tuple) Contains 3 columns - date_time, price, and volume.
        :return: (list) Extracted bars
        """

        # Iterate over rows
        list_bars = []

        for row in data:
            # Set variables
            date_time = row[0]
            self.tick_num += 1
            price = float(row[1])
            volume = row[2]
            dollar_value = price * volume
            signed_tick = self._apply_tick_rule(price)

            if isinstance(self.threshold, (int, float)):
                # If the threshold is fixed, it's used for every sampling
                threshold = self.threshold
            else:
                # If the threshold is changing, then the threshold defined just before
                # sampling time is used
                threshold = self.threshold.iloc[self.threshold.index.get_indexer([date_time], method='pad')[0]]

            if self.open_price is None:
                self.open_price = price

            # Update high low prices
            self.high_price, self.low_price = self._update_high_low(price)

            # Calculations
            self.cum_statistics['cum_ticks'] += 1
            self.cum_statistics['cum_dollar_value'] += dollar_value
            self.cum_statistics['cum_volume'] += volume
            if signed_tick == 1:
                self.cum_statistics['cum_buy_volume'] += volume

            # If threshold reached then take a sample
            if self.cum_statistics[self.metric] >= threshold:  # pylint: disable=eval-used
                self._create_bars(date_time, price,
                                  self.high_price, self.low_price, list_bars)

                # Reset cache
                self._reset_cache()
        return list_bars


def get_dollar_bars(file_path_or_df: Union[str, Iterable[str], pd.DataFrame], threshold: Union[float, pd.Series] = 70000000,
                    batch_size: int = 20000000, verbose: bool = True, to_csv: bool = False, output_path: Optional[str] = None):
    """
    Creates the dollar bars: date_time, open, high, low, close, volume, cum_buy_volume, cum_ticks, cum_dollar_value.

    Following the paper "The Volume Clock: Insights into the high frequency paradigm" by Lopez de Prado, et al,
    it is suggested that using 1/50 of the average daily dollar value, would result in more desirable statistical
    properties.

    :param file_path_or_df: (str, iterable of str, or pd.DataFrame) Path to the csv file(s) or Pandas Data Frame containing raw tick data
                            in the format[date_time, price, volume]
    :param threshold: (float, or pd.Series) A cumulative value above this threshold triggers a sample to be taken.
                      If a series is given, then at each sampling time the closest previous threshold is used.
                      (Values in the series can only be at times when the threshold is changed, not for every observation)
    :param batch_size: (int) The number of rows per batch. Less RAM = smaller batch size.
    :param verbose: (bool) Print out batch numbers (True or False)
    :param to_csv: (bool) Save bars to csv after every batch run (True or False)
    :param output_path: (str) Path to csv file, if to_csv is True
    :return: (pd.DataFrame) Dataframe of dollar bars
    """

    bars = StandardBars(metric='cum_dollar_value', threshold=threshold, batch_size=batch_size)
    dollar_bars = bars.batch_run(file_path_or_df=file_path_or_df, verbose=verbose, to_csv=to_csv, output_path=output_path)
    return dollar_bars


def get_volume_bars(file_path_or_df: Union[str, Iterable[str], pd.DataFrame], threshold: Union[float, pd.Series] = 70000000,
                    batch_size: int = 20000000, verbose: bool = True, to_csv: bool = False, output_path: Optional[str] = None):
    """
    Creates the volume bars: date_time, open, high, low, close, volume, cum_buy_volume, cum_ticks, cum_dollar_value.

    Following the paper "The Volume Clock: Insights into the high frequency paradigm" by Lopez de Prado, et al,
    it is suggested that using 1/50 of the average daily volume, would result in more desirable statistical properties.

    :param file_path_or_df: (str, iterable of str, or pd.DataFrame) Path to the csv file(s) or Pandas Data Frame containing raw tick data
                            in the format[date_time, price, volume]
    :param threshold: (float, or pd.Series) A cumulative value above this threshold triggers a sample to be taken.
                      If a series is given, then at each sampling time the closest previous threshold is used.
                      (Values in the series can only be at times when the threshold is changed, not for every observation)
    :param batch_size: (int) The number of rows per batch. Less RAM = smaller batch size.
    :param verbose: (bool) Print out batch numbers (True or False)
    :param to_csv: (bool) Save bars to csv after every batch run (True or False)
    :param output_path: (str) Path to csv file, if to_csv is True
    :return: (pd.DataFrame) Dataframe of volume bars
    """
    bars = StandardBars(metric='cum_volume', threshold=threshold, batch_size=batch_size)
    volume_bars = bars.batch_run(file_path_or_df=file_path_or_df, verbose=verbose, to_csv=to_csv, output_path=output_path)
    return volume_bars


def get_tick_bars(file_path_or_df: Union[str, Iterable[str], pd.DataFrame], threshold: Union[float, pd.Series] = 70000000,
                  batch_size: int = 20000000, verbose: bool = True, to_csv: bool = False, output_path: Optional[str] = None):
    """
    Creates the tick bars: date_time, open, high, low, close, volume, cum_buy_volume, cum_ticks, cum_dollar_value.

    :param file_path_or_df: (str, iterable of str, or pd.DataFrame) Path to the csv file(s) or Pandas Data Frame containing raw tick data
                             in the format[date_time, price, volume]
    :param threshold: (float, or pd.Series) A cumulative value above this threshold triggers a sample to be taken.
                      If a series is given, then at each sampling time the closest previous threshold is used.
                      (Values in the series can only be at times when the threshold is changed, not for every observation)
    :param batch_size: (int) The number of rows per batch. Less RAM = smaller batch size.
    :param verbose: (bool) Print out batch numbers (True or False)
    :param to_csv: (bool) Save bars to csv after every batch run (True or False)
    :param output_path: (str) Path to csv file, if to_csv is True
    :return: (pd.DataFrame) Dataframe of volume bars
    """
    bars = StandardBars(metric='cum_ticks',
                        threshold=threshold, batch_size=batch_size)
    tick_bars = bars.batch_run(file_path_or_df=file_path_or_df, verbose=verbose, to_csv=to_csv, output_path=output_path)
    return tick_bars



"""
Chapter 3.2 Fixed-Time Horizon Method, in Advances in Financial Machine Learning, by M. L. de Prado.

Work "Classification-based Financial Markets Prediction using Deep Neural Networks" by Dixon et al. (2016) describes how
labeling data this way can be used in training deep neural networks to predict price movements.
"""

import warnings
import pandas as pd


def fixed_time_horizon(prices, threshold=0, resample_by=None, lag=True, standardized=False, window=None):
    """
    Fixed-Time Horizon Labeling Method.

    Originally described in the book Advances in Financial Machine Learning, Chapter 3.2, p.43-44.

    Returns 1 if return is greater than the threshold, -1 if less, and 0 if in between. If no threshold is
    provided then it will simply take the sign of the return.

    :param prices: (pd.Series or pd.DataFrame) Time-indexed stock prices used to calculate returns.
    :param threshold: (float or pd.Series) When the absolute value of return exceeds the threshold, the observation is
                    labeled with 1 or -1, depending on the sign of the return. If return is less, it's labeled as 0.
                    Can be dynamic if threshold is inputted as a pd.Series, and threshold.index must match prices.index.
                    If resampling is used, the index of threshold must match the index of prices after resampling.
                    If threshold is negative, then the directionality of the labels will be reversed. If no threshold
                    is provided, it is assumed to be 0 and the sign of the return is returned.
    :param resample_by: (str) If not None, the resampling period for price data prior to calculating returns. 'B' = per
                        business day, 'W' = week, 'M' = month, etc. Will take the last observation for each period.
                        For full details see `here.
                        <https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects>`_
    :param lag: (bool) If True, returns will be lagged to make them forward-looking.
    :param standardized: (bool) Whether returns are scaled by mean and standard deviation.
    :param window: (int) If standardized is True, the rolling window period for calculating the mean and standard
                    deviation of returns.
    :return: (pd.Series or pd.DataFrame) -1, 0, or 1 denoting whether the return for each observation is
                    less/between/greater than the threshold at each corresponding time index. First or last row will be
                    NaN, depending on lag.
    """
    # Apply resample period, if applicable.
    if resample_by is not None:
        prices = prices.resample(resample_by).last()

    # Calculate returns.
    if lag:
        returns = prices.pct_change(1).shift(-1)
    else:
        returns = prices.pct_change(1)

    # If threshold is pd.Series, its index must patch prices.index; otherwise labels will fail to return.
    if isinstance(threshold, pd.Series):
        assert threshold.index.equals(prices.index), "prices.index and threshold.index must match! If prices are " \
                                                     "resampled, the threshold index must match the resampled prices " \
                                                     "index."

    # Adjust by mean and stdev, if desired. Assert that window must exist if standardization is on. Warning if window
    # is too large.
    if standardized:
        assert isinstance(window, int), "When standardized is True, window must be int."
        if window >= len(returns):
            warnings.warn('window is greater than the length of the Series. All labels will be NaN.', UserWarning)

        # Apply standardization.
        mean = returns.rolling(window=window).mean()
        stdev = returns.rolling(window=window).std()
        returns -= mean
        returns /= stdev

    # Apply labeling.
    labels = returns.copy()  # Copy returns so labels aren't all 0 when threshold => 1.
    labels[returns.lt(-threshold, axis=0)] = -1
    labels[returns.gt(threshold, axis=0)] = 1
    labels[(returns.ge(-threshold, axis=0)) & (returns.le(threshold, axis=0))] = 0

    return labels



# pylint: disable=missing-module-docstring
import numpy as np
from famlafl.online_portfolio_selection.base import OLPS


class OLMAR(OLPS):
    """
    This class implements the Online Moving Average Reversion strategy. It is reproduced with
    modification from the following paper:
    `Li, Bin & Hoi, Steven. (2012). On-Line Portfolio
    Selection with Moving Average Reversion. Proceedings of the 29th International Conference on
    Machine Learning, ICML 2012. 1. <https://arxiv.org/pdf/1206.4626.pdf>`_

    Online Moving Average Reversion reverts to the SMA or EMA of the underlying assets based on
    the given threshold.
    """

    def __init__(self, reversion_method, epsilon, window=None, alpha=None):
        """
        Initializes Online Moving Average Reversion with the given reversion method, epsilon,
        window, and alpha.

        :param reversion_method: (int) 1 for SMA, 2 for EWA. Typically 2 has higher returns than 1,
                                       but this is dependant on the dataset.
        :param epsilon: (float) Reversion threshold with range [1, inf). OLMAR methods typically do
                                not rely on epsilon values as it is more dependant on the window and
                                alpha value; however, epsilon of 20 seems to be a safe choice to
                                avoid extreme values.
        :param window: (int) Number of windows to calculate Simple Moving Average with range [1, inf).
                             This parameter depends on the data. If the data has a shorter mean reversion
                             trends, then window of 6 can have high returns. If the data has a longer
                             mean reversion trend, then values close to 21 can have high returns.
        :param alpha: (float) Exponential ratio for Exponentially Weighted Average with range (0, 1).
                              Larger alpha value indicates more importance placed on recent prices,
                              and smaller alpha value indicates more importance placed on the past.
        """
        self.epsilon = epsilon
        self.window = window
        self.alpha = alpha
        self.reversion_method = reversion_method
        self.moving_average_reversion = None  # (np.array) Calculated moving average.
        super().__init__()

    def _initialize(self, asset_prices, weights, resample_by):
        """
        Initializes the important variables for the object.

        :param asset_prices: (pd.DataFrame) Historical asset prices.
        :param weights: (list/np.array/pd.DataFrame) Initial weights set by the user.
        :param resample_by: (str) Specifies how to resample the prices. 'D' for Day, 'W' for Week,
                                 'M' for Month. The inputs are based on pandas' resample method.
        """
        super(OLMAR, self)._initialize(asset_prices, weights, resample_by)

        # Check that reversion method is either 1 or 2.
        if self.reversion_method not in [1, 2]:
            raise ValueError("Reversion method must be either 1 or 2.")

        # Check that epsilon values are correct.
        if self.epsilon < 1:
            raise ValueError("Epsilon values must be greater or equal to 1.")

        # Check that window is at least 1.
        if self.reversion_method == 1 and self.window < 1:
            raise ValueError("Window must be at least 1.")

        # Chat that alpha is between 0 and 1 for method 2.
        if self.reversion_method == 2 and (self.alpha < 0 or self.alpha > 1):
            raise ValueError("Alpha must be between 0 and 1.")

        # Calculate moving average reversion.
        self.moving_average_reversion = self._calculate_rolling_moving_average()

    def _update_weight(self, time):
        """
        Predicts the next time's portfolio weight.

        :param time: (int) Current time period.
        :return: (np.array) Predicted weights.
        """
        # Return predetermined weights for time periods with no significant data.
        if self.reversion_method == 1 and time < self.window or time == 0:
            return self.weights

        # Get predicted change through SMA or EWA.
        predicted_change = self.moving_average_reversion[time]

        # Calculate the mean of the predicted change.
        mean_relative = np.mean(predicted_change)

        # Portfolio weights of mean prediction.
        mean_change = np.ones(self.number_of_assets) * mean_relative

        # Loss function to switch mean reversion strategy.
        loss_fn = max(0, (self.epsilon - np.dot(self.weights, predicted_change)))

        # If loss function is 0, set multiplicative constant to zero.
        if loss_fn == 0:
            lambd = 0
        # If not, adjust lambda, a multiplicative constant.
        else:
            check = np.linalg.norm(predicted_change - mean_change) ** 2
            if check != 0:
                lambd = loss_fn / check
            else:
                lambd = 0
        new_weights = self.weights + lambd * (predicted_change - mean_change)

        # Projects new weights to simplex domain.
        new_weights = self._simplex_projection(new_weights)
        return new_weights

    def _calculate_rolling_moving_average(self):
        """
        Calculates the rolling moving average for Online Moving Average Reversion.

        :return: (np.array) Rolling moving average for the given reversion method.
        """
        # MAR-1 reversion method: Simple Moving Average.
        if self.reversion_method == 1:
            rolling_ma = np.array(self.asset_prices.rolling(self.window).apply(
                lambda x: np.sum(x) / x[0] / self.window, raw=True))
        # MAR-2 reversion method: Exponential Moving Average.
        else:
            rolling_ma = np.array(
                self.asset_prices.ewm(alpha=self.alpha, adjust=False).mean() / self.asset_prices)
        return rolling_ma



"""
Tests the financial data structures
"""

import unittest
import os
import numpy as np
import pandas as pd

from famlafl.data_structures import imbalance_data_structures as ds


class TestDataStructures(unittest.TestCase):
    """
    Test the various financial data structures:
    1. Imbalance Dollar bars
    2. Imbalance Volume bars
    3. Imbalance Tick bars
    """

    def setUp(self):
        """
        Set the file path for the tick data csv
        """
        project_path = os.path.dirname(__file__)
        self.path = project_path + '/test_data/imbalance_sample_data_small.csv'

    def test_ema_imbalance_dollar_bars(self):
        """
        Tests the EMA imbalance dollar bars implementation.
        """
        exp_num_ticks_init = 1000
        num_prev_bars = 3

        db1, _ = ds.get_ema_dollar_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                  expected_imbalance_window=10000,
                                                  num_prev_bars=num_prev_bars, batch_size=2e7, verbose=False, )
        db2, _ = ds.get_ema_dollar_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                  expected_imbalance_window=10000,
                                                  num_prev_bars=num_prev_bars, batch_size=50, verbose=False)
        db3, _ = ds.get_ema_dollar_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                  expected_imbalance_window=10000,
                                                  num_prev_bars=num_prev_bars, batch_size=10, verbose=False)
        ds.get_ema_dollar_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                         expected_imbalance_window=10000,
                                         num_prev_bars=num_prev_bars, batch_size=50, verbose=False,
                                         to_csv=True, output_path='test.csv')
        db4 = pd.read_csv('test.csv', parse_dates=[0])

        self.assertEqual(db1.shape, (624, 10))
        #self.assertEqual(db1.shape, (4770, 10))

        # Assert diff batch sizes have same number of bars
        self.assertTrue(db1.shape == db2.shape)
        self.assertTrue(db1.shape == db3.shape)
        self.assertTrue(db1.shape == db4.shape)

        # Assert same values
        self.assertTrue(np.all(db1.values == db2.values))
        self.assertTrue(np.all(db1.values == db3.values))
        self.assertTrue(np.all(db1.values == db4.values))

        # Assert OHLC is correct (the first value)
        self.assertEqual(db1.loc[0, 'open'], 1306.0)
        self.assertEqual(db1.loc[0, 'high'], 1306.0)
        self.assertEqual(db1.loc[0, 'low'], 1304.25)
        self.assertEqual(db1.loc[0, 'close'], 1304.5)

        self.assertTrue((db1.loc[:, 'high'] >= db1.loc[:, 'low']).all())

        # Assert OHLC is correct (some index)
        self.assertEqual(db1.loc[600, 'open'], 1304.5)
        self.assertEqual(db1.loc[600, 'high'], 1304.5)
        self.assertEqual(db1.loc[600, 'low'], 1304.5)
        self.assertEqual(db1.loc[600, 'close'], 1304.5)

        self.assertTrue((db1.loc[:, 'volume'] >= db1.loc[:, 'cum_buy_volume']).all())

        # Delete generated csv file (if it wasn't generated test would fail)
        os.remove('test.csv')

    def test_ema_imbalance_volume_bars(self):
        """
        Tests the EMA imbalance volume bars implementation.
        """
        exp_num_ticks_init = 1000
        num_prev_bars = 3

        db1, _ = ds.get_ema_volume_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                  expected_imbalance_window=10000,
                                                  num_prev_bars=num_prev_bars, batch_size=2e7, verbose=False, )
        db2, _ = ds.get_ema_volume_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                  expected_imbalance_window=10000,
                                                  num_prev_bars=num_prev_bars, batch_size=50, verbose=False)
        db3, _ = ds.get_ema_volume_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                  expected_imbalance_window=10000,
                                                  num_prev_bars=num_prev_bars, batch_size=10, verbose=False)
        ds.get_ema_volume_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                         expected_imbalance_window=10000,
                                         num_prev_bars=num_prev_bars, batch_size=50, verbose=False,
                                         to_csv=True, output_path='test.csv')
        db4 = pd.read_csv('test.csv', parse_dates=[0])

        self.assertEqual(db1.shape, (624, 10))

        # Assert diff batch sizes have same number of bars
        self.assertTrue(db1.shape == db2.shape)
        self.assertTrue(db1.shape == db3.shape)
        self.assertTrue(db1.shape == db4.shape)

        # Assert same values
        self.assertTrue(np.all(db1.values == db2.values))
        self.assertTrue(np.all(db1.values == db3.values))
        self.assertTrue(np.all(db1.values == db4.values))

        # Assert OHLC is correct (the first value)
        self.assertEqual(db1.loc[0, 'open'], 1306.0)
        self.assertEqual(db1.loc[0, 'high'], 1306.0)
        self.assertEqual(db1.loc[0, 'low'], 1304.25)
        self.assertEqual(db1.loc[0, 'close'], 1304.5)

        # Assert OHLC is correct (some index)
        self.assertEqual(db1.loc[600, 'open'], 1304.5)
        self.assertEqual(db1.loc[600, 'high'], 1304.5)
        self.assertEqual(db1.loc[600, 'low'], 1304.5)
        self.assertEqual(db1.loc[600, 'close'], 1304.5)
        self.assertTrue((db1.loc[:, 'high'] >= db1.loc[:, 'low']).all())
        self.assertTrue((db1.loc[:, 'volume'] >= db1.loc[:, 'cum_buy_volume']).all())

        # Delete generated csv file (if it wasn't generated test would fail)
        os.remove('test.csv')

    def test_ema_imbalance_tick_bars(self):
        """
        Tests the EMA imbalance tick bars implementation.
        """
        exp_num_ticks_init = 1000
        num_prev_bars = 3

        db1, _ = ds.get_ema_tick_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                expected_imbalance_window=10000,
                                                num_prev_bars=num_prev_bars, batch_size=2e7, verbose=False, )
        db2, _ = ds.get_ema_tick_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                expected_imbalance_window=10000,
                                                num_prev_bars=num_prev_bars, batch_size=50, verbose=False)
        db3, _ = ds.get_ema_tick_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                expected_imbalance_window=10000,
                                                num_prev_bars=num_prev_bars, batch_size=10, verbose=False)
        ds.get_ema_tick_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                       expected_imbalance_window=10000,
                                       num_prev_bars=num_prev_bars, batch_size=50, verbose=False,
                                       to_csv=True, output_path='test.csv')
        db4 = pd.read_csv('test.csv', parse_dates=[0])

        self.assertEqual(db1.shape, (3558, 10))

        # Assert diff batch sizes have same number of bars
        self.assertTrue(db1.shape == db2.shape)
        self.assertTrue(db1.shape == db3.shape)
        self.assertTrue(db1.shape == db4.shape)

        # Assert same values
        self.assertTrue(np.all(db1.values == db2.values))
        self.assertTrue(np.all(db1.values == db3.values))
        self.assertTrue(np.all(db1.values == db4.values))

        # Assert OHLC is correct (the first value)
        self.assertEqual(db1.loc[0, 'open'], 1306.0)
        self.assertEqual(db1.loc[0, 'high'], 1306.0)
        self.assertEqual(db1.loc[0, 'low'], 1304.25)
        self.assertEqual(db1.loc[0, 'close'], 1304.25)

        # Assert OHLC is correct (some index)
        self.assertEqual(db1.loc[600, 'open'], 1306.75)
        self.assertEqual(db1.loc[600, 'high'], 1306.75)
        self.assertEqual(db1.loc[600, 'low'], 1306.75)
        self.assertEqual(db1.loc[600, 'close'], 1306.75)
        self.assertTrue((db1.loc[:, 'high'] >= db1.loc[:, 'low']).all())
        self.assertTrue((db1.loc[:, 'volume'] >= db1.loc[:, 'cum_buy_volume']).all())

        # Delete generated csv file (if it wasn't generated test would fail)
        os.remove('test.csv')

    def test_ema_imb_dollar_bars_with_constraints(self):
        """
        Test the EMA Dollar Imbalance bars with expected number of ticks max and min constraints
        """
        exp_num_ticks_init = 1000
        num_prev_bars = 3
        exp_num_ticks_constraints = [100, 1000]

        db1, _ = ds.get_ema_dollar_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                  expected_imbalance_window=10000,
                                                  exp_num_ticks_constraints=exp_num_ticks_constraints,
                                                  num_prev_bars=num_prev_bars, batch_size=2e7, verbose=False)
        db2, _ = ds.get_ema_dollar_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                  expected_imbalance_window=10000,
                                                  exp_num_ticks_constraints=exp_num_ticks_constraints,
                                                  num_prev_bars=num_prev_bars, batch_size=50, verbose=False)
        db3, _ = ds.get_ema_dollar_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                  expected_imbalance_window=10000,
                                                  exp_num_ticks_constraints=exp_num_ticks_constraints,
                                                  num_prev_bars=num_prev_bars, batch_size=10, verbose=False)
        ds.get_ema_dollar_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                         expected_imbalance_window=10000,
                                         exp_num_ticks_constraints=exp_num_ticks_constraints,
                                         num_prev_bars=num_prev_bars, batch_size=50, verbose=False,
                                         to_csv=True, output_path='test.csv')
        db4 = pd.read_csv('test.csv', parse_dates=[0])

        self.assertEqual(db1.shape, (583, 10))

        # Assert diff batch sizes have same number of bars
        self.assertTrue(db1.shape == db2.shape)
        self.assertTrue(db1.shape == db3.shape)
        self.assertTrue(db1.shape == db4.shape)

        # Assert same values
        self.assertTrue(np.all(db1.values == db2.values))
        self.assertTrue(np.all(db1.values == db3.values))
        self.assertTrue(np.all(db1.values == db4.values))

        # Assert OHLC is correct (the first value)
        self.assertEqual(db1.loc[0, 'open'], 1306.0)
        self.assertEqual(db1.loc[0, 'high'], 1306.0)
        self.assertEqual(db1.loc[0, 'low'], 1304.25)
        self.assertEqual(db1.loc[0, 'close'], 1304.5)

        self.assertTrue((db1.loc[:, 'high'] >= db1.loc[:, 'low']).all())

        # Assert OHLC is correct (some index)
        self.assertEqual(db1.loc[500, 'open'], 1303.5)
        self.assertEqual(db1.loc[500, 'high'], 1303.5)
        self.assertEqual(db1.loc[500, 'low'], 1303.5)
        self.assertEqual(db1.loc[500, 'close'], 1303.5)

        self.assertTrue((db1.loc[:, 'volume'] >= db1.loc[:, 'cum_buy_volume']).all())

        # Delete generated csv file (if it wasn't generated test would fail)
        os.remove('test.csv')

    def test_const_imbalance_dollar_bars(self):
        """
        Tests the Const imbalance dollar bars implementation.
        """
        exp_num_ticks_init = 1000

        db1, _ = ds.get_const_dollar_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                    expected_imbalance_window=10000,
                                                    batch_size=2e7, verbose=False, )
        db2, _ = ds.get_const_dollar_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                    expected_imbalance_window=10000,
                                                    batch_size=50, verbose=False)
        db3, _ = ds.get_const_dollar_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                    expected_imbalance_window=10000,
                                                    batch_size=10, verbose=False)
        ds.get_const_dollar_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                           expected_imbalance_window=10000,
                                           batch_size=50, verbose=False,
                                           to_csv=True, output_path='test.csv')
        db4 = pd.read_csv('test.csv', parse_dates=[0])

        self.assertEqual(db1.shape, (109, 10))

        # Assert diff batch sizes have same number of bars
        self.assertTrue(db1.shape == db2.shape)
        self.assertTrue(db1.shape == db3.shape)
        self.assertTrue(db1.shape == db4.shape)

        # Assert same values
        self.assertTrue(np.all(db1.values == db2.values))
        self.assertTrue(np.all(db1.values == db3.values))
        self.assertTrue(np.all(db1.values == db4.values))

        # Assert OHLC is correct (the first value)
        self.assertEqual(db1.loc[0, 'open'], 1306.0)
        self.assertEqual(db1.loc[0, 'high'], 1306.0)
        self.assertEqual(db1.loc[0, 'low'], 1304.25)
        self.assertEqual(db1.loc[0, 'close'], 1304.5)

        self.assertTrue((db1.loc[:, 'high'] >= db1.loc[:, 'low']).all())

        # Assert OHLC is correct (some index)
        self.assertEqual(db1.loc[105, 'open'], 1304.5)
        self.assertEqual(db1.loc[105, 'high'], 1304.75)
        self.assertEqual(db1.loc[105, 'low'], 1304.25)
        self.assertEqual(db1.loc[105, 'close'], 1304.75)

        self.assertTrue((db1.loc[:, 'volume'] >= db1.loc[:, 'cum_buy_volume']).all())

        # Delete generated csv file (if it wasn't generated test would fail)
        os.remove('test.csv')

    def test_const_imbalance_volume_bars(self):
        """
        Tests the Const imbalance volume bars implementation.
        """
        exp_num_ticks_init = 1000

        db1, thresh1 = ds.get_const_volume_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                          expected_imbalance_window=10000,
                                                          batch_size=2e7, verbose=False, analyse_thresholds=True)
        db2, thresh2 = ds.get_const_volume_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                          expected_imbalance_window=10000,
                                                          batch_size=50, verbose=False, analyse_thresholds=True)
        db3, _ = ds.get_const_volume_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                    expected_imbalance_window=10000,
                                                    batch_size=10, verbose=False)
        ds.get_const_volume_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                           expected_imbalance_window=10000,
                                           batch_size=50, verbose=False,
                                           to_csv=True, output_path='test.csv')
        db4 = pd.read_csv('test.csv', parse_dates=[0])

        self.assertEqual(db1.shape, (112, 10))

        self.assertTrue(np.all(thresh1.cum_theta == thresh2.cum_theta))

        # Assert diff batch sizes have same number of bars
        self.assertTrue(db1.shape == db2.shape)
        self.assertTrue(db1.shape == db3.shape)
        self.assertTrue(db1.shape == db4.shape)

        # Assert same values
        self.assertTrue(np.all(db1.values == db2.values))
        self.assertTrue(np.all(db1.values == db3.values))
        self.assertTrue(np.all(db1.values == db4.values))

        # Assert OHLC is correct (the first value)
        self.assertEqual(db1.loc[0, 'open'], 1306.0)
        self.assertEqual(db1.loc[0, 'high'], 1306.0)
        self.assertEqual(db1.loc[0, 'low'], 1304.25)
        self.assertEqual(db1.loc[0, 'close'], 1304.5)

        self.assertTrue((db1.loc[:, 'high'] >= db1.loc[:, 'low']).all())

        # Assert OHLC is correct (some index)
        self.assertEqual(db1.loc[100, 'open'], 1303.75)
        self.assertEqual(db1.loc[100, 'high'], 1303.75)
        self.assertEqual(db1.loc[100, 'low'], 1303.75)
        self.assertEqual(db1.loc[100, 'close'], 1303.75)

        self.assertTrue((db1.loc[:, 'volume'] >= db1.loc[:, 'cum_buy_volume']).all())

        # Delete generated csv file (if it wasn't generated test would fail)
        os.remove('test.csv')

    def test_const_imbalance_tick_bars(self):
        """
        Tests the Const imbalance tick bars implementation.
        """
        exp_num_ticks_init = 1000

        db1, _ = ds.get_const_tick_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                  expected_imbalance_window=10000,
                                                  batch_size=2e7, verbose=False)
        db2, _ = ds.get_const_tick_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                  expected_imbalance_window=10000,
                                                  batch_size=50, verbose=False)
        db3, _ = ds.get_const_tick_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                                  expected_imbalance_window=10000,
                                                  batch_size=10, verbose=False)
        ds.get_const_tick_imbalance_bars(self.path, exp_num_ticks_init=exp_num_ticks_init,
                                         expected_imbalance_window=10000,
                                         batch_size=50, verbose=False,
                                         to_csv=True, output_path='test.csv')
        db4 = pd.read_csv('test.csv', parse_dates=[0])

        self.assertEqual(db1.shape, (55, 10))

        # Assert diff batch sizes have same number of bars
        self.assertTrue(db1.shape == db2.shape)
        self.assertTrue(db1.shape == db3.shape)
        self.assertTrue(db1.shape == db4.shape)

        # Assert same values
        self.assertTrue(np.all(db1.values == db2.values))
        self.assertTrue(np.all(db1.values == db3.values))
        self.assertTrue(np.all(db1.values == db4.values))

        # Assert OHLC is correct (the first value)
        self.assertEqual(db1.loc[0, 'open'], 1306.0)
        self.assertEqual(db1.loc[0, 'high'], 1306.0)
        self.assertEqual(db1.loc[0, 'low'], 1304.25)
        self.assertEqual(db1.loc[0, 'close'], 1304.25)

        self.assertTrue((db1.loc[:, 'high'] >= db1.loc[:, 'low']).all())

        # Assert OHLC is correct (some index)
        self.assertEqual(db1.loc[50, 'open'], 1305.25)
        self.assertEqual(db1.loc[50, 'high'], 1305.25)
        self.assertEqual(db1.loc[50, 'low'], 1305)
        self.assertEqual(db1.loc[50, 'close'], 1305)

        self.assertTrue((db1.loc[:, 'volume'] >= db1.loc[:, 'cum_buy_volume']).all())

        # Delete generated csv file (if it wasn't generated test would fail)
        os.remove('test.csv')

    def test_csv_format(self):
        """
        Asserts that the csv data being passed is of the correct format.
        """
        wrong_date = ['2019-41-30', 200.00, np.int64(5)]
        wrong_price = ['2019-01-30', 'asd', np.int64(5)]
        wrong_volume = ['2019-01-30', 200.00, '1.5']
        too_many_cols = ['2019-01-30', 200.00,
                         np.int64(5), 'Limit order', 'B23']

        # pylint: disable=protected-access
        self.assertRaises(ValueError, ds.BaseImbalanceBars._assert_csv,
                          pd.DataFrame(wrong_date).T)
        # pylint: disable=protected-access
        self.assertRaises(AssertionError,
                          ds.BaseImbalanceBars._assert_csv,
                          pd.DataFrame(too_many_cols).T)
        # pylint: disable=protected-access
        self.assertRaises(AssertionError,
                          ds.BaseImbalanceBars._assert_csv,
                          pd.DataFrame(wrong_price).T)
        # pylint: disable=protected-access
        self.assertRaises(AssertionError,
                          ds.BaseImbalanceBars._assert_csv,
                          pd.DataFrame(wrong_volume).T)

    def test_wrong_imbalance_passed(self):
        """
        Tests ValueError raise when wrong imbalance was passed
        """
        bar_gen = ds.ConstImbalanceBars(metric='cum_buy_volume', expected_imbalance_window=10,
                                        exp_num_ticks_init=100,
                                        analyse_thresholds=False, batch_size=10000)
        with self.assertRaises(ValueError):
            bar_gen.batch_run(self.path, verbose=False)



"""
Tests different algorithms in the Mean Variance class.
"""

import unittest
from unittest.mock import patch
from io import StringIO
import os
import numpy as np
import pandas as pd
from famlafl.portfolio_optimization.mean_variance import MeanVarianceOptimisation
from famlafl.portfolio_optimization.returns_estimators import ReturnsEstimators


class TestMVO(unittest.TestCase):
    # pylint: disable=too-many-public-methods
    """
    Tests the different functions of the Mean Variance Optimisation class
    """

    def setUp(self):
        """
        Set the file path for the tick data csv.
        """

        project_path = os.path.dirname(__file__)
        data_path = project_path + '/test_data/stock_prices.csv'
        self.data = pd.read_csv(data_path, parse_dates=True, index_col="Date")

    def test_inverse_variance_solution(self):
        """
        Test the calculation of inverse-variance portfolio weights.
        """

        mvo = MeanVarianceOptimisation()
        mvo.allocate(asset_prices=self.data, solution='inverse_variance', asset_names=self.data.columns)
        weights = mvo.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_min_volatility_solution(self):
        """
        Test the calculation of minimum volatility portfolio weights.
        """

        mvo = MeanVarianceOptimisation()
        mvo.allocate(asset_prices=self.data, solution='min_volatility', asset_names=self.data.columns)
        weights = mvo.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

        # Check that the volatility is the minimum among all other portfolios
        for solution_string in {"inverse_variance", "max_sharpe", "max_return_min_volatility",
                                "max_diversification", "max_decorrelation"}:
            mvo_ = MeanVarianceOptimisation()
            mvo_.allocate(asset_prices=self.data, solution=solution_string, asset_names=self.data.columns)
            assert mvo.portfolio_risk <= mvo_.portfolio_risk

    def test_max_return_min_volatility_solution(self):
        """
        Test the calculation of maximum expected return and minimum volatility portfolio weights.
        """

        mvo = MeanVarianceOptimisation()
        mvo.allocate(asset_prices=self.data,
                     risk_aversion=50,
                     solution='max_return_min_volatility',
                     asset_names=self.data.columns)
        weights = mvo.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_max_sharpe_solution(self):
        """
        Test the calculation of maximum sharpe portfolio weights.
        """

        mvo = MeanVarianceOptimisation()
        mvo.allocate(asset_prices=self.data, solution='max_sharpe', asset_names=self.data.columns)
        weights = mvo.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_min_volatility_for_target_return(self):
        # pylint: disable=invalid-name
        """
        Test the calculation of minimum volatility-target return portfolio weights.
        """

        mvo = MeanVarianceOptimisation()
        prices = self.data.resample('W').last()
        mvo.allocate(asset_prices=prices, solution='efficient_risk', asset_names=self.data.columns)
        weights = mvo.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_max_return_for_target_risk(self):
        # pylint: disable=invalid-name
        """
        Test the calculation of maximum return-target risk portfolio weights.
        """

        mvo = MeanVarianceOptimisation()
        mvo.allocate(asset_prices=self.data, solution='efficient_return', asset_names=self.data.columns)
        weights = mvo.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_max_diversification(self):
        # pylint: disable=invalid-name
        """
        Test the calculation of maximum diversification portfolio weights.
        """

        mvo = MeanVarianceOptimisation()
        mvo.allocate(asset_prices=self.data, solution='max_diversification', asset_names=self.data.columns)
        weights = mvo.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_max_decorrelation(self):
        # pylint: disable=invalid-name
        """
        Test the calculation of maximum decorrelation portfolio weights.
        """

        mvo = MeanVarianceOptimisation()
        mvo.allocate(asset_prices=self.data, solution='max_decorrelation', asset_names=self.data.columns)
        weights = mvo.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_plotting_efficient_frontier(self):
        # pylint: disable=invalid-name, protected-access
        """
        Test the plotting of the efficient frontier.
        """

        mvo = MeanVarianceOptimisation()
        expected_returns = ReturnsEstimators().calculate_mean_historical_returns(asset_prices=self.data,
                                                                                 resample_by='W')
        covariance = ReturnsEstimators().calculate_returns(asset_prices=self.data, resample_by='W').cov()
        plot = mvo.plot_efficient_frontier(covariance=covariance,
                                           max_return=1.0,
                                           expected_asset_returns=expected_returns)
        assert plot.axes.xaxis.label._text == 'Volatility'
        assert plot.axes.yaxis.label._text == 'Return'
        assert len(plot._A) == 41

    def test_exception_in_plotting_efficient_frontier(self):
        # pylint: disable=invalid-name, protected-access
        """
        Test raising of exception when plotting the efficient frontier.
        """

        mvo = MeanVarianceOptimisation()
        expected_returns = ReturnsEstimators().calculate_mean_historical_returns(asset_prices=self.data,
                                                                                 resample_by='W')
        covariance = ReturnsEstimators().calculate_returns(asset_prices=self.data, resample_by='W').cov()
        plot = mvo.plot_efficient_frontier(covariance=covariance,
                                           max_return=1.0,
                                           expected_asset_returns=expected_returns)
        assert len(plot._A) == 41

    def test_mvo_with_input_as_returns_and_covariance(self):
        # pylint: disable=invalid-name
        """
        Test MVO when we pass expected returns and covariance matrix as input.
        """

        mvo = MeanVarianceOptimisation()
        expected_returns = ReturnsEstimators().calculate_mean_historical_returns(asset_prices=self.data, resample_by='W')
        covariance = ReturnsEstimators().calculate_returns(asset_prices=self.data, resample_by='W').cov()
        mvo.allocate(covariance_matrix=covariance,
                     expected_asset_returns=expected_returns,
                     asset_names=self.data.columns)
        weights = mvo.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_min_volatility_with_specific_weight_bounds(self):
        # pylint: disable=invalid-name
        """
        Test the calculation of weights when specific bounds are supplied.
        """

        mvo = MeanVarianceOptimisation()
        mvo.allocate(asset_prices=self.data,
                     solution='min_volatility',
                     weight_bounds=['weights[0] <= 0.2'],
                     asset_names=self.data.columns)
        weights = mvo.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_max_sharpe_with_specific_weight_bounds(self):
        # pylint: disable=invalid-name
        """
        Test the calculation of weights when specific bounds are supplied.
        """

        mvo = MeanVarianceOptimisation()
        mvo.allocate(asset_prices=self.data,
                     solution='max_sharpe',
                     weight_bounds=['y[0] <= kappa * 0.5'],
                     asset_names=self.data.columns)
        weights = mvo.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_efficient_risk_with_specific_weight_bounds(self):
        # pylint: disable=invalid-name
        """
        Test the calculation of weights when specific bounds are supplied.
        """

        mvo = MeanVarianceOptimisation()
        mvo.allocate(asset_prices=self.data,
                     solution='efficient_risk',
                     target_return=0.01,
                     weight_bounds=['weights[0] <= 0.3'],
                     asset_names=self.data.columns)
        weights = mvo.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_efficient_return_with_specific_weight_bounds(self):
        # pylint: disable=invalid-name
        """
        Test the calculation of weights when specific bounds are supplied.
        """

        mvo = MeanVarianceOptimisation()
        mvo.allocate(asset_prices=self.data,
                     solution='efficient_return',
                     weight_bounds=['weights[0] <= 0.3'],
                     asset_names=self.data.columns)
        weights = mvo.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_max_decorrelation_with_specific_weight_bounds(self):
        # pylint: disable=invalid-name
        """
        Test the calculation of weights when specific bounds are supplied.
        """

        mvo = MeanVarianceOptimisation()
        mvo.allocate(asset_prices=self.data,
                     solution='max_decorrelation',
                     weight_bounds=['weights[0] <= 0.3'],
                     asset_names=self.data.columns)
        weights = mvo.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_max_diversification_with_specific_weight_bounds(self):
        # pylint: disable=invalid-name
        """
        Test the calculation of weights when specific bounds are supplied.
        """

        mvo = MeanVarianceOptimisation()
        mvo.allocate(asset_prices=self.data,
                     solution='max_diversification',
                     weight_bounds=['weights[0] <= 0.3'],
                     asset_names=self.data.columns)
        weights = mvo.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_max_return_min_volatility_with_specific_weight_bounds(self):
        # pylint: disable=invalid-name
        """
        Test the calculation of weights when specific bounds are supplied.
        """

        mvo = MeanVarianceOptimisation()
        mvo.allocate(asset_prices=self.data,
                     solution='max_return_min_volatility',
                     weight_bounds=['weights[0] <= 0.3'],
                     asset_names=self.data.columns)
        weights = mvo.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_mvo_with_exponential_returns(self):
        # pylint: disable=invalid-name
        """
        Test the calculation of inverse-variance portfolio weights.
        """

        mvo = MeanVarianceOptimisation(calculate_expected_returns='exponential')
        mvo.allocate(asset_prices=self.data, solution='inverse_variance', asset_names=self.data.columns)
        weights = mvo.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_unknown_returns_calculation(self):
        # pylint: disable=invalid-name
        """
        Test ValueError on passing unknown returns calculation string.
        """

        with self.assertRaises(ValueError):
            mvo = MeanVarianceOptimisation(calculate_expected_returns='unknown_returns')
            mvo.allocate(asset_prices=self.data, asset_names=self.data.columns)

    def test_value_error_for_unknown_solution(self):
        # pylint: disable=invalid-name
        """
        Test ValueError on passing unknown solution string.
        """

        with self.assertRaises(ValueError):
            mvo = MeanVarianceOptimisation()
            mvo.allocate(asset_prices=self.data, solution='ivp', asset_names=self.data.columns)

    def test_value_error_for_non_dataframe_input(self):
        # pylint: disable=invalid-name
        """
        Test ValueError on passing non-dataframe input.
        """

        with self.assertRaises(ValueError):
            mvo = MeanVarianceOptimisation()
            mvo.allocate(asset_prices=self.data.values, solution='inverse_variance', asset_names=self.data.columns)

    def test_value_error_for_no_min_volatility_optimal_weights(self):
        # pylint: disable=invalid-name
        """
        Test ValueError when no optimal weights are found for minimum volatility solution.
        """

        with self.assertRaises(ValueError):
            mvo = MeanVarianceOptimisation()
            mvo.allocate(asset_prices=self.data,
                         solution='min_volatility',
                         weight_bounds=(0.9, 1),
                         asset_names=self.data.columns)

    def test_value_error_for_no_quadratic_utlity_optimal_weights(self):
        # pylint: disable=invalid-name
        """
        Test ValueError when no optimal weights are found for max return-minimum volatility solution.
        """

        with self.assertRaises(ValueError):
            mvo = MeanVarianceOptimisation()
            mvo.allocate(asset_prices=self.data,
                         solution='max_return_min_volatility',
                         weight_bounds=(0.9, 1),
                         asset_names=self.data.columns)

    def test_value_error_for_no_max_sharpe_optimal_weights(self):
        # pylint: disable=invalid-name
        """
        Test ValueError when no optimal weights are found for maximum Sharpe solution.
        """

        with self.assertRaises(ValueError):
            mvo = MeanVarianceOptimisation()
            mvo.allocate(asset_prices=self.data,
                         solution='max_sharpe',
                         weight_bounds=(0.9, 1),
                         asset_names=self.data.columns)

    def test_value_error_for_no_efficient_risk_optimal_weights(self):
        # pylint: disable=invalid-name
        """
        Test ValueError when no optimal weights are found for efficient risk solution.
        """

        with self.assertRaises(ValueError):
            mvo = MeanVarianceOptimisation()
            mvo.allocate(asset_prices=self.data,
                         solution='efficient_risk',
                         weight_bounds=(0.9, 1),
                         asset_names=self.data.columns)

    def test_value_error_for_no_efficient_return_optimal_weights(self):
        # pylint: disable=invalid-name
        """
        Test ValueError when no optimal weights are found for efficient return solution.
        """

        with self.assertRaises(ValueError):
            mvo = MeanVarianceOptimisation()
            mvo.allocate(asset_prices=self.data,
                         solution='efficient_return',
                         weight_bounds=(0.9, 1),
                         asset_names=self.data.columns)

    def test_value_error_for_no_max_diversification_optimal_weights(self):
        # pylint: disable=invalid-name
        """
        Test ValueError when no optimal weights are found for max diversification solution.
        """

        with self.assertRaises(ValueError):
            mvo = MeanVarianceOptimisation()
            mvo.allocate(asset_prices=self.data,
                         solution='max_diversification',
                         weight_bounds=(0.9, 1),
                         asset_names=self.data.columns)

    def test_value_error_for_no_max_decorrelation_optimal_weights(self):
        # pylint: disable=invalid-name
        """
        Test ValueError when no optimal weights are found for max decorrelation solution.
        """

        with self.assertRaises(ValueError):
            mvo = MeanVarianceOptimisation()
            mvo.allocate(asset_prices=self.data,
                         solution='max_decorrelation',
                         weight_bounds=(0.9, 1),
                         asset_names=self.data.columns)

    def test_value_error_for_non_date_index(self):
        """
        Test ValueError on passing dataframe not indexed by date.
        """

        with self.assertRaises(ValueError):
            mvo = MeanVarianceOptimisation()
            data = self.data.reset_index()
            mvo.allocate(asset_prices=data, solution='inverse_variance', asset_names=self.data.columns)

    def test_resampling_asset_prices(self):
        """
        Test resampling of asset prices.
        """

        mvo = MeanVarianceOptimisation()
        mvo.allocate(asset_prices=self.data, solution='inverse_variance', asset_names=self.data.columns)
        weights = mvo.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_all_inputs_none(self):
        """
        Test allocation when all inputs are None.
        """

        with self.assertRaises(ValueError):
            mvo = MeanVarianceOptimisation()
            mvo.allocate(asset_names=self.data.columns)

    def test_no_asset_names(self):
        """
        Test MVO when not supplying a list of asset names.
        """

        mvo = MeanVarianceOptimisation()
        mvo.allocate(asset_prices=self.data)
        weights = mvo.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_no_asset_names_by_passing_cov(self):
        """
        Test MVO when not supplying a list of asset names but passing covariance matrix as input
        """

        mvo = MeanVarianceOptimisation()
        expected_returns = ReturnsEstimators().calculate_exponential_historical_returns(asset_prices=self.data,
                                                                                        resample_by='W')
        covariance = ReturnsEstimators().calculate_returns(asset_prices=self.data, resample_by='W').cov()
        mvo.allocate(expected_asset_returns=expected_returns, covariance_matrix=covariance)
        weights = mvo.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_valuerror_with_no_asset_names(self):
        """
        Test ValueError when not supplying a list of asset names and no other input
        """

        with self.assertRaises(ValueError):
            mvo = MeanVarianceOptimisation()
            expected_returns = ReturnsEstimators().calculate_mean_historical_returns(asset_prices=self.data,
                                                                                     resample_by='W')
            covariance = ReturnsEstimators().calculate_returns(asset_prices=self.data, resample_by='W').cov()
            mvo.allocate(expected_asset_returns=expected_returns, covariance_matrix=covariance.values)

    def test_portfolio_metrics(self):
        """
        Test the printing of portfolio metrics to stdout.
        """

        mvo = MeanVarianceOptimisation()
        mvo.allocate(asset_prices=self.data)
        with patch('sys.stdout', new=StringIO()) as fake_out:
            mvo.get_portfolio_metrics()
            output = fake_out.getvalue().strip()
            self.assertTrue('Portfolio Return = 0.017362404155484328' in output)
            self.assertTrue('Portfolio Risk = 9.385801639141577e-06' in output)
            self.assertTrue('Portfolio Sharpe Ratio = 9.385801639141577e-06' in output)

    def test_custom_objective_function(self):
        """
        Test custom portfolio objective and allocation constraints.
        """

        mvo = MeanVarianceOptimisation()
        custom_obj = 'cp.Minimize(risk)'
        constraints = ['cp.sum(weights) == 1', 'weights >= 0', 'weights <= 1']
        non_cvxpy_variables = {
            'num_assets': self.data.shape[1],
            'covariance': self.data.cov(),
            'expected_returns': ReturnsEstimators().calculate_mean_historical_returns(asset_prices=self.data,
                                                                                      resample_by='W')
        }
        cvxpy_variables = [
            'risk = cp.quad_form(weights, covariance)',
            'portfolio_return = cp.matmul(weights, expected_returns)'
        ]
        mvo.allocate_custom_objective(non_cvxpy_variables=non_cvxpy_variables,
                                      cvxpy_variables=cvxpy_variables,
                                      objective_function=custom_obj,
                                      constraints=constraints)
        weights = mvo.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        assert mvo.asset_names == list(range(mvo.num_assets))
        assert mvo.portfolio_return == 0.012854555899642236
        assert  mvo.portfolio_risk == 3.0340907720046832
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_custom_objective_with_asset_names(self):
        """
        Test custom portfolio objective and constraints while providing a list of asset names.
        """

        mvo = MeanVarianceOptimisation()
        custom_obj = 'cp.Minimize(kappa)'
        constraints = ['cp.sum(weights) == 1', 'weights >= 0', 'weights <= 1']
        non_cvxpy_variables = {
            'num_assets': self.data.shape[1],
            'covariance': self.data.cov(),
            'asset_names': self.data.columns
        }
        cvxpy_variables = [
            'kappa = cp.quad_form(weights, covariance)',
        ]
        mvo.allocate_custom_objective(
            non_cvxpy_variables=non_cvxpy_variables,
            cvxpy_variables=cvxpy_variables,
            objective_function=custom_obj,
            constraints=constraints)
        weights = mvo.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        assert list(mvo.asset_names) == list(self.data.columns)
        assert mvo.portfolio_return is None
        assert mvo.portfolio_risk is None
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_value_error_for_custom_obj_optimal_weights(self):
        # pylint: disable=invalid-name
        """
        Test ValueError when no optimal weights are found for custom objective solution.
        """

        with self.assertRaises(ValueError):
            mvo = MeanVarianceOptimisation()
            custom_obj = 'cp.Minimize(risk - kappa)'
            constraints = ['cp.sum(weights) == 1', 'weights >= 0', 'weights <= 1']
            non_cvxpy_variables = {
                'num_assets': self.data.shape[1],
                'covariance': self.data.cov(),
                'expected_returns': ReturnsEstimators().calculate_mean_historical_returns(asset_prices=self.data,
                                                                                          resample_by='W')
            }
            cvxpy_variables = [
                'risk = cp.quad_form(weights, covariance)',
                'portfolio_return = cp.matmul(weights, expected_returns)',
                'kappa = cp.Variable(1)'
            ]
            mvo.allocate_custom_objective(non_cvxpy_variables=non_cvxpy_variables,
                                          cvxpy_variables=cvxpy_variables,
                                          objective_function=custom_obj,
                                          constraints=constraints)



"""
Tests the different portfolio optimisation algorithms
"""

import unittest
import os
import numpy as np
import pandas as pd
from famlafl.portfolio_optimization.herc import HierarchicalEqualRiskContribution
from famlafl.portfolio_optimization.returns_estimators import ReturnsEstimators


class TestHERC(unittest.TestCase):
    # pylint: disable=too-many-public-methods
    """
    Tests different functions of the HERC algorithm class.
    """

    def setUp(self):
        """
        Set the file path for the tick data csv
        """
        project_path = os.path.dirname(__file__)
        data_path = project_path + '/test_data/stock_prices.csv'
        self.data = pd.read_csv(data_path, parse_dates=True, index_col="Date")

    def test_herc_equal_weight(self):
        """
        Test the weights calculated by the HERC algorithm - if all the weights are positive and
        their sum is equal to 1.
        """

        herc = HierarchicalEqualRiskContribution()
        herc.allocate(asset_prices=self.data,
                      asset_names=self.data.columns,
                      optimal_num_clusters=5,
                      risk_measure='equal_weighting')
        weights = herc.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_herc_min_variance(self):
        """
        Test the weights calculated by the HERC algorithm - if all the weights are positive and
        their sum is equal to 1.
        """

        herc = HierarchicalEqualRiskContribution()
        herc.allocate(asset_prices=self.data,
                      asset_names=self.data.columns,
                      optimal_num_clusters=5,
                      risk_measure='variance')
        weights = herc.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_herc_min_standard_deviation(self):
        """
        Test the weights calculated by the HERC algorithm - if all the weights are positive and
        their sum is equal to 1.
        """

        herc = HierarchicalEqualRiskContribution()
        herc.allocate(asset_prices=self.data,
                      asset_names=self.data.columns,
                      optimal_num_clusters=4,
                      risk_measure='standard_deviation')
        weights = herc.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_value_error_for_expected_shortfall(self):
        """
        Test ValueError when expected_shortfall is the allocation metric, no asset_returns dataframe
        is given and no asset_prices dataframe is passed.
        """

        with self.assertRaises(ValueError):
            herc = HierarchicalEqualRiskContribution()
            herc.allocate(asset_names=self.data.columns,
                          optimal_num_clusters=5,
                          risk_measure='expected_shortfall')

    def test_herc_expected_shortfall(self):
        """
        Test the weights calculated by the HERC algorithm - if all the weights are positive and
        their sum is equal to 1.
        """

        herc = HierarchicalEqualRiskContribution()
        herc.allocate(asset_prices=self.data,
                      asset_names=self.data.columns,
                      optimal_num_clusters=5,
                      risk_measure='expected_shortfall')
        weights = herc.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_herc_conditional_drawdown_risk(self):
        """
        Test the weights calculated by the HERC algorithm - if all the weights are positive and
        their sum is equal to 1.
        """

        herc = HierarchicalEqualRiskContribution()
        herc.allocate(asset_prices=self.data,
                      asset_names=self.data.columns,
                      optimal_num_clusters=5,
                      risk_measure='conditional_drawdown_risk')
        weights = herc.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_quasi_diagnalization(self):
        """
        Test the quasi-diagnalisation step of HERC algorithm.
        """

        herc = HierarchicalEqualRiskContribution()
        herc.allocate(asset_prices=self.data,
                      linkage='single',
                      optimal_num_clusters=5,
                      asset_names=self.data.columns)
        assert herc.ordered_indices == [13, 9, 10, 8, 14, 7, 1, 6, 4, 16, 3, 17,
                                        12, 18, 22, 0, 15, 21, 11, 2, 20, 5, 19]

    def test_value_error_for_non_dataframe_input(self):
        """
        Test ValueError on passing non-dataframe input.
        """

        with self.assertRaises(ValueError):
            herc = HierarchicalEqualRiskContribution()
            herc.allocate(asset_prices=self.data.values, asset_names=self.data.columns)

    def test_value_error_for_non_date_index(self):
        """
        Test ValueError on passing dataframe not indexed by date.
        """

        with self.assertRaises(ValueError):
            herc = HierarchicalEqualRiskContribution()
            data = self.data.reset_index()
            herc.allocate(asset_prices=data, asset_names=self.data.columns)

    def test_all_inputs_none(self):
        """
        Test allocation when all inputs are None.
        """

        with self.assertRaises(ValueError):
            herc = HierarchicalEqualRiskContribution()
            herc.allocate(asset_names=self.data.columns)

    def test_herc_with_input_as_returns(self):
        """
        Test HERC when passing asset returns dataframe as input.
        """

        herc = HierarchicalEqualRiskContribution()
        returns = ReturnsEstimators().calculate_returns(asset_prices=self.data)
        herc.allocate(asset_returns=returns, asset_names=self.data.columns)
        weights = herc.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_herc_with_asset_returns_as_none(self):
        """
        Test HERC when asset returns are not required for calculating the weights.
        """

        herc = HierarchicalEqualRiskContribution()
        returns = ReturnsEstimators().calculate_returns(asset_prices=self.data)
        herc.allocate(asset_names=self.data.columns,
                      covariance_matrix=returns.cov(),
                      optimal_num_clusters=5,
                      risk_measure='equal_weighting')
        weights = herc.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_herc_with_input_as_covariance_matrix(self):
        """
        Test HERC when passing a covariance matrix as input.
        """

        herc = HierarchicalEqualRiskContribution()
        returns = ReturnsEstimators().calculate_returns(asset_prices=self.data)
        herc.allocate(asset_names=self.data.columns,
                      covariance_matrix=returns.cov(),
                      optimal_num_clusters=6,
                      asset_returns=returns)
        weights = herc.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_value_error_for_risk_measure(self):
        """
        Test HERC when a different allocation metric string is used.
        """

        with self.assertRaises(ValueError):
            herc = HierarchicalEqualRiskContribution()
            herc.allocate(asset_names=self.data.columns, asset_prices=self.data, risk_measure='random_metric')

    def test_no_asset_names(self):
        """
        Test HERC when not supplying a list of asset names.
        """

        herc = HierarchicalEqualRiskContribution()
        herc.allocate(asset_prices=self.data,
                      optimal_num_clusters=6)
        weights = herc.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_no_asset_names_with_asset_returns(self):
        """
        Test HERC when not supplying a list of asset names and when the user passes asset_returns.
        """

        herc = HierarchicalEqualRiskContribution()
        returns = ReturnsEstimators().calculate_returns(asset_prices=self.data)
        herc.allocate(asset_returns=returns,
                      optimal_num_clusters=6)
        weights = herc.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_value_error_with_no_asset_names(self):
        """
        Test ValueError when not supplying a list of asset names and no other input
        """

        with self.assertRaises(ValueError):
            herc = HierarchicalEqualRiskContribution()
            returns = ReturnsEstimators().calculate_returns(asset_prices=self.data)
            herc.allocate(asset_returns=returns.values,
                          optimal_num_clusters=6)

    def test_dendrogram_plot(self):
        """
        Test if dendrogram plot object is correctly rendered.
        """

        herc = HierarchicalEqualRiskContribution()
        herc.allocate(asset_prices=self.data, optimal_num_clusters=5)
        dendrogram = herc.plot_clusters(assets=self.data.columns)
        assert dendrogram.get('icoord')
        assert dendrogram.get('dcoord')
        assert dendrogram.get('ivl')
        assert dendrogram.get('leaves')
        assert dendrogram.get('color_list')



"""
Test various functions regarding chapter 4: Sampling (Bootstrapping, Concurrency).
"""

import os
import unittest

import numpy as np
import pandas as pd

from sklearn.metrics import precision_score, roc_auc_score, accuracy_score, mean_absolute_error, \
    mean_squared_error
from sklearn.model_selection import TimeSeriesSplit
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import LinearSVC

from famlafl.util.volatility import get_daily_vol
from famlafl.filters.filters import cusum_filter
from famlafl.labeling.labeling import get_events, add_vertical_barrier, get_bins
from famlafl.sampling.bootstrapping import get_ind_matrix, get_ind_mat_label_uniqueness
from famlafl.ensemble.sb_bagging import SequentiallyBootstrappedBaggingClassifier, \
    SequentiallyBootstrappedBaggingRegressor


# pylint: disable=invalid-name

def _generate_label_with_prob(x, prob, random_state=np.random.RandomState(1)):
    """
    Generates true label value with some probability(prob)
    """
    choice = random_state.choice([0, 1], p=[1 - prob, prob])
    if choice == 1:
        return x
    return int(not x)


def _get_synthetic_samples(ind_mat, good_samples_thresh, bad_samples_thresh):
    """
    Get samples with uniqueness either > good_samples_thresh or uniqueness < bad_samples_thresh
    """
    print("ind_mat shape:", ind_mat.shape)
    if ind_mat.size == 0:
        print("Warning: ind_mat is empty")
        return list(range(100))  # Return first 100 samples as fallback

    # Get mix of samples where some of them are extremely non-overlapping, the other one are highly overlapping
    i = 0
    unique_samples = []
    uniqueness_values = []
    label_uniqueness = get_ind_mat_label_uniqueness(ind_mat)
    print("label_uniqueness shape:", label_uniqueness.shape)
    
    for label in label_uniqueness:
        if label.size > 0:
            mean_uniqueness = np.mean(label[label > 0]) if np.any(label > 0) else 0
            uniqueness_values.append(mean_uniqueness)
            if mean_uniqueness > good_samples_thresh or mean_uniqueness < bad_samples_thresh:
                unique_samples.append(i)
        i += 1
    
    print(f"Initial unique samples: {len(unique_samples)}")
    print(f"Uniqueness values stats: min={min(uniqueness_values) if uniqueness_values else 0}, "
          f"max={max(uniqueness_values) if uniqueness_values else 0}, "
          f"mean={np.mean(uniqueness_values) if uniqueness_values else 0}")
    
    if len(unique_samples) < 100:  # Ensure we have at least 100 samples
        print("Not enough samples, using percentile-based selection")
        uniqueness_array = np.array(uniqueness_values)
        if len(uniqueness_array) > 0:
            # Get samples from both tails of the distribution
            low_thresh = np.percentile(uniqueness_array, 25)
            high_thresh = np.percentile(uniqueness_array, 75)
            unique_samples = []
            for i, val in enumerate(uniqueness_values):
                if val <= low_thresh or val >= high_thresh:
                    unique_samples.append(i)
    
    if len(unique_samples) == 0:
        print("Still no samples, using first 100 indices")
        unique_samples = list(range(min(100, ind_mat.shape[0] if ind_mat.size > 0 else 100)))
    
    print("Final number of unique samples:", len(unique_samples))
    return unique_samples


class TestSequentiallyBootstrappedBagging(unittest.TestCase):
    """
    Test SequentiallyBootstrapped Bagging classifiers
    """

    def setUp(self):
        """
        Set the file path for the sample dollar bars data and get triple barrier events, generate features
        """
        project_path = os.path.dirname(__file__)
        self.path = project_path + '/test_data/dollar_bar_sample.csv'
        self.data = pd.read_csv(self.path, index_col='date_time')
        self.data.index = pd.to_datetime(self.data.index)

        print("\nInitial data shape:", self.data.shape)

        # Compute moving averages
        self.data['fast_mavg'] = self.data['close'].rolling(window=20, min_periods=20,
                                                            center=False).mean()
        self.data['slow_mavg'] = self.data['close'].rolling(window=50, min_periods=50,
                                                            center=False).mean()

        # Compute sides
        self.data['side'] = np.nan

        long_signals = self.data['fast_mavg'] >= self.data['slow_mavg']
        short_signals = self.data['fast_mavg'] < self.data['slow_mavg']
        self.data.loc[long_signals, 'side'] = 1
        self.data.loc[short_signals, 'side'] = -1

        # Remove Look ahead bias by lagging the signal
        self.data['side'] = self.data['side'].shift(1)

        daily_vol = get_daily_vol(close=self.data['close'], lookback=50) * 2  # Increase target size
        cusum_events = cusum_filter(self.data['close'], threshold=0.001)  # Further lower threshold
        print("cusum_events length:", len(cusum_events))
        
        vertical_barriers = add_vertical_barrier(t_events=cusum_events, close=self.data['close'],
                                                 num_hours=2)
        print("vertical_barriers length:", len(vertical_barriers))
        
        meta_labeled_events = get_events(close=self.data['close'],
                                         t_events=cusum_events,
                                         pt_sl=[1, 1],  # More aggressive profit taking/stop loss
                                         target=daily_vol,
                                         min_ret=1e-6,  # Further lower minimum return threshold
                                         num_threads=3,
                                         vertical_barrier_times=vertical_barriers,
                                         side_prediction=self.data['side'],
                                         verbose=False)
        meta_labeled_events.dropna(inplace=True)
        print("meta_labeled_events shape:", meta_labeled_events.shape)
        
        labels = get_bins(meta_labeled_events, self.data['close'])
        print("labels shape:", labels.shape)

        # Generate data set which shows the power of SB Bagging vs Standard Bagging
        print("meta_labeled_events.t1 shape:", meta_labeled_events.t1.shape)
        print("meta_labeled_events.t1 head:\n", meta_labeled_events.t1.head())
        print("self.data.close shape:", self.data.close.shape)
        print("self.data.close head:\n", self.data.close.head())
        
        ind_mat = get_ind_matrix(meta_labeled_events.t1, self.data.close)
        print("ind_mat shape:", ind_mat.shape)

        # Use more relaxed thresholds to ensure we get samples
        unique_samples = _get_synthetic_samples(ind_mat, 0.3, 0.05)

        # Get synthetic data set with drawn samples
        print("Shape before synthetic samples:", self.data.shape)
        print("Length of unique_samples:", len(unique_samples))
        X = self.data.loc[labels.index].iloc[unique_samples].copy()  # Use copy() to avoid SettingWithCopyWarning
        print("Shape after initial X creation:", X.shape)
        X = X.dropna()
        print("Shape after first dropna:", X.shape)
        labels = labels.loc[X.index, :]
        print("Shape of labels:", labels.shape)
        X['y'] = labels.bin  # Add y column directly
        print("Shape after adding y column:", X.shape)

        # Generate features (some of them are informative, others are just noise)
        if 'y' not in X.columns:
            raise ValueError("Column 'y' is missing from DataFrame X")
            
        # Create probability columns all at once using vectorized operations
        for prob in [0.6, 0.5, 0.3, 0.2, 0.1]:
            col_name = f'label_prob_{prob}'
            X[col_name] = X['y'].apply(lambda x: _generate_label_with_prob(x, prob))

        # Ensure all label probability columns exist before creating rolling means
        required_columns = ['label_prob_0.6', 'label_prob_0.5', 'label_prob_0.3', 'label_prob_0.2', 'label_prob_0.1']
        for col in required_columns:
            if col not in X.columns:
                raise ValueError(f"Required column {col} is missing from DataFrame")

        features = ['label_prob_0.6', 'label_prob_0.2']  # Two super-informative features
        for prob in [0.5, 0.3, 0.2, 0.1]:
            for window in [2, 5, 10]:
                X['label_prob_{}_sma_{}'.format(prob, window)] = X['label_prob_{}'.format(prob)].rolling(
                    window=window).mean()
                features.append('label_prob_{}_sma_{}'.format(prob, window))
        print("Shape before final dropna:", X.shape)
        X.dropna(inplace=True)
        print("Shape after final dropna:", X.shape)
        y = X.pop('y')
        print("Length of y:", len(y))

        # Create TimeSeriesSplit cross-validator with fewer splits and minimum size
        n_samples = len(X)
        if n_samples > 0:
            test_size = max(2, int(n_samples * 0.2))  # Ensure at least 2 samples in test
            tscv = TimeSeriesSplit(n_splits=2, test_size=test_size, gap=0)
            
            # Get the first split for train/test
            for train_indices, test_indices in tscv.split(X[features]):
                self.X_train = X[features].iloc[train_indices]
                self.X_test = X[features].iloc[test_indices]
                self.y_train_clf = y.iloc[train_indices]
                self.y_test_clf = y.iloc[test_indices]
                break  # Only use first split
        else:
            raise ValueError("No samples available for train/test split")

        self.y_train_reg = (1 + self.y_train_clf)
        self.y_test_reg = (1 + self.y_test_clf)

        self.samples_info_sets = meta_labeled_events.loc[self.X_train.index, 't1']
        self.price_bars_trim = self.data[
            (self.data.index >= self.X_train.index.min()) & (self.data.index <= self.X_train.index.max())].close

    def test_sb_bagging_not_tree_base_estimator(self):
        """
        Test SB Bagging with non-tree base estimator (KNN)
        """
        clf = KNeighborsClassifier()
        sb_clf = SequentiallyBootstrappedBaggingClassifier(base_estimator=clf,
                                                           samples_info_sets=self.samples_info_sets,
                                                           price_bars=self.price_bars_trim)
        sb_clf.fit(self.X_train, self.y_train_clf)
        self.assertTrue((sb_clf.predict(self.X_train)[:10] == np.array([0, 0, 0, 0, 0, 0, 1, 1, 1, 0])).all)

    def test_sb_bagging_non_sample_weights_with_verbose(self):
        """
        Test SB Bagging with classifier which doesn't support sample_weights with verbose > 1
        """
        clf = LinearSVC()

        sb_clf = SequentiallyBootstrappedBaggingClassifier(base_estimator=clf, max_features=0.2,
                                                           n_estimators=2,
                                                           samples_info_sets=self.samples_info_sets,
                                                           price_bars=self.price_bars_trim, oob_score=True,
                                                           random_state=1, bootstrap_features=True,
                                                           max_samples=30, verbose=2)
        with self.assertWarns(UserWarning):
            sb_clf.fit(self.X_train, self.y_train_clf)
        self.assertTrue((sb_clf.predict(self.X_train)[:10] == np.array([0, 0, 0, 0, 0, 0, 1, 1, 1, 0])).all)

    def test_sb_bagging_with_max_features(self):
        """
        Test SB Bagging with base_estimator bootstrap = True, float max_features, max_features bootstrap = True
        """
        clf = RandomForestClassifier(n_estimators=1, criterion='entropy', bootstrap=True,
                                     class_weight='balanced_subsample', max_depth=12)

        sb_clf = SequentiallyBootstrappedBaggingClassifier(base_estimator=clf, max_features=0.2,
                                                           n_estimators=2,
                                                           samples_info_sets=self.samples_info_sets,
                                                           price_bars=self.price_bars_trim, oob_score=True,
                                                           random_state=1, bootstrap_features=True,
                                                           max_samples=30)
        with self.assertWarns(UserWarning):
            sb_clf.fit(self.X_train, self.y_train_clf, sample_weight=np.ones((self.X_train.shape[0],)))
        self.assertTrue((sb_clf.predict(self.X_train)[:10] == np.array([0, 0, 0, 0, 0, 0, 1, 1, 1, 0])).all)

    def test_sb_bagging_float_max_samples_warm_start_true(self):
        """
        Test SB Bagging with warm start = True and float max_samples
        """
        clf = RandomForestClassifier(n_estimators=1, criterion='entropy', bootstrap=False,
                                     class_weight='balanced_subsample', max_depth=12)

        sb_clf = SequentiallyBootstrappedBaggingClassifier(base_estimator=clf, max_features=7,
                                                           n_estimators=2,
                                                           samples_info_sets=self.samples_info_sets,
                                                           price_bars=self.price_bars_trim, oob_score=False,
                                                           random_state=1, bootstrap_features=True,
                                                           max_samples=0.3, warm_start=True)

        sb_clf.fit(self.X_train, self.y_train_clf, sample_weight=np.ones((self.X_train.shape[0],)))

        sb_clf.n_estimators += 0
        with self.assertWarns(UserWarning):
            sb_clf.fit(self.X_train, self.y_train_clf, sample_weight=np.ones((self.X_train.shape[0],)))
        sb_clf.n_estimators += 2
        sb_clf.fit(self.X_train, self.y_train_clf, sample_weight=np.ones((self.X_train.shape[0],)))

        self.assertTrue((sb_clf.predict(self.X_train)[:10] == np.array([0, 0, 0, 0, 0, 0, 1, 1, 1, 0])).all)

    def test_value_error_raise(self):
        """
        Test various values error raise
        """
        clf = KNeighborsClassifier()
        bagging_clf_1 = SequentiallyBootstrappedBaggingClassifier(base_estimator=clf,
                                                                  samples_info_sets=self.samples_info_sets,
                                                                  price_bars=self.data)
        bagging_clf_2 = SequentiallyBootstrappedBaggingClassifier(base_estimator=clf,
                                                                  samples_info_sets=self.samples_info_sets,
                                                                  price_bars=self.data, max_samples=2000000)
        bagging_clf_3 = SequentiallyBootstrappedBaggingClassifier(base_estimator=clf,
                                                                  samples_info_sets=self.samples_info_sets,
                                                                  price_bars=self.data, max_features='20')
        bagging_clf_4 = SequentiallyBootstrappedBaggingClassifier(base_estimator=clf,
                                                                  samples_info_sets=self.samples_info_sets,
                                                                  price_bars=self.data, max_features=2000000)
        bagging_clf_5 = SequentiallyBootstrappedBaggingClassifier(base_estimator=clf,
                                                                  samples_info_sets=self.samples_info_sets,
                                                                  price_bars=self.data, oob_score=True, warm_start=True)
        bagging_clf_6 = SequentiallyBootstrappedBaggingClassifier(base_estimator=clf,
                                                                  samples_info_sets=self.samples_info_sets,
                                                                  price_bars=self.data, warm_start=True)
        bagging_clf_7 = SequentiallyBootstrappedBaggingClassifier(base_estimator=clf,
                                                                  samples_info_sets=self.samples_info_sets,
                                                                  price_bars=self.data, warm_start=True)
        with self.assertRaises(ValueError):
            # ValueError to use sample weight with classifier which doesn't support sample weights
            bagging_clf_1.fit(self.X_train, self.y_train_clf, sample_weight=np.ones((self.X_train.shape[0],)))
        with self.assertRaises(ValueError):
            # ValueError for max_samples > X_train.shape[0]
            bagging_clf_2.fit(self.X_train, self.y_train_clf,
                              sample_weight=np.ones((self.X_train.shape[0],)))
        with self.assertRaises(ValueError):
            # ValueError for non-int/float max_features param
            bagging_clf_3.fit(self.X_train, self.y_train_clf,
                              sample_weight=np.ones((self.X_train.shape[0],)))
        with self.assertRaises(ValueError):
            # ValueError for max_features > X_train.shape[1]
            bagging_clf_4.fit(self.X_train, self.y_train_clf,
                              sample_weight=np.ones((self.X_train.shape[0],)))
        with self.assertRaises(ValueError):
            # ValueError for warm_start and oob_score being True
            bagging_clf_5.fit(self.X_train, self.y_train_clf,
                              sample_weight=np.ones((self.X_train.shape[0],)))
        with self.assertRaises(ValueError):
            # ValueError for decreasing the number of estimators when warm start is True
            bagging_clf_6.fit(self.X_train, self.y_train_clf)
            bagging_clf_6.n_estimators -= 2
            bagging_clf_6.fit(self.X_train, self.y_train_clf)
        with self.assertRaises(ValueError):
            # ValueError for setting n_estimators to negative value
            bagging_clf_7.fit(self.X_train, self.y_train_clf)
            bagging_clf_7.n_estimators -= 1000
            bagging_clf_7.fit(self.X_train, self.y_train_clf)

    def test_sb_classifier(self):
        """
        Test Sequentially Bootstrapped Bagging Classifier. Here we compare oos/oob scores to sklearn's bagging oos scores,
        test oos predictions values
        """
        # Init classifiers
        clf_base = RandomForestClassifier(n_estimators=1, criterion='entropy', bootstrap=False,
                                          class_weight='balanced_subsample')

        sb_clf = SequentiallyBootstrappedBaggingClassifier(base_estimator=clf_base, max_features=1.0, n_estimators=100,
                                                           samples_info_sets=self.samples_info_sets,
                                                           price_bars=self.price_bars_trim, oob_score=True,
                                                           random_state=1)

        # X_train index should be in index mapping
        self.assertTrue(self.X_train.index.isin(sb_clf.timestamp_int_index_mapping.index).all())

        sb_clf.fit(self.X_train, self.y_train_clf)

        self.assertTrue((sb_clf.X_time_index == self.X_train.index).all())  # X_train index == clf X_train index

        oos_sb_predictions = sb_clf.predict(self.X_test)

        sb_precision = precision_score(self.y_test_clf, oos_sb_predictions)
        sb_roc_auc = roc_auc_score(self.y_test_clf, oos_sb_predictions)
        sb_accuracy = accuracy_score(self.y_test_clf, oos_sb_predictions)

        self.assertAlmostEqual(sb_accuracy, 0.66, delta=0.2)
        self.assertEqual(sb_precision, 1.0)
        self.assertAlmostEqual(sb_roc_auc, 0.59, delta=0.2)

    def test_sb_regressor(self):
        """
        Test Sequentially Bootstrapped Bagging Regressor
        """
        # Init regressors
        reg = RandomForestRegressor(n_estimators=1, bootstrap=False)
        sb_reg = SequentiallyBootstrappedBaggingRegressor(base_estimator=reg, max_features=1.0, n_estimators=100,
                                                          samples_info_sets=self.samples_info_sets,
                                                          price_bars=self.price_bars_trim, oob_score=True,
                                                          random_state=1)
        sb_reg_1 = SequentiallyBootstrappedBaggingRegressor(base_estimator=reg, max_features=1.0, n_estimators=1,
                                                            samples_info_sets=self.samples_info_sets,
                                                            price_bars=self.price_bars_trim, oob_score=True,
                                                            random_state=1)

        sb_reg.fit(self.X_train, self.y_train_reg)

        with self.assertWarns(UserWarning):
            sb_reg_1.fit(self.X_train, self.y_train_reg)  # To raise warning and get code coverage

        # X_train index should be in index mapping
        self.assertTrue(self.X_train.index.isin(sb_reg.timestamp_int_index_mapping.index).all())
        self.assertTrue((sb_reg.X_time_index == self.X_train.index).all())  # X_train index == reg X_train index

        oos_sb_predictions = sb_reg.predict(self.X_test)
        mse_sb_reg = mean_squared_error(self.y_test_reg, oos_sb_predictions)
        mae_sb_reg = mean_absolute_error(self.y_test_reg, oos_sb_predictions)

        self.assertAlmostEqual(mse_sb_reg, 0.16, delta=0.1)
        self.assertAlmostEqual(mae_sb_reg, 0.29, delta=0.1)



"""
Tests the Hierarchical Risk Parity (HRP) algorithm.
"""

import unittest
import os
import numpy as np
import pandas as pd
from famlafl.portfolio_optimization.hrp import HierarchicalRiskParity
from famlafl.portfolio_optimization.returns_estimators import ReturnsEstimators


class TestHRP(unittest.TestCase):
    # pylint: disable=too-many-public-methods
    """
    Tests different functions of the HRP algorithm class.
    """

    def setUp(self):
        """
        Set the file path for the tick data csv.
        """

        project_path = os.path.dirname(__file__)
        data_path = project_path + '/test_data/stock_prices.csv'
        self.data = pd.read_csv(data_path, parse_dates=True, index_col="Date")

    def test_hrp(self):
        """
        Test the weights calculated by the HRP algorithm - if all the weights are positive and
        their sum is equal to 1.
        """

        hrp = HierarchicalRiskParity()
        hrp.allocate(asset_prices=self.data, asset_names=self.data.columns)
        weights = hrp.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_hrp_long_short(self):
        """
        Test the Long Short Portfolio via side_weights Serries 1 for Long, -1 for Short (index=asset names)
        """
        hrp = HierarchicalRiskParity()
        side_weights = pd.Series([1] * self.data.shape[1], index=self.data.columns)
        side_weights.loc[self.data.columns[:4]] = -1
        hrp.allocate(asset_prices=self.data, asset_names=self.data.columns, side_weights=side_weights)
        weights = hrp.weights.values[0]
        self.assertEqual(len(weights) - self.data.shape[1], 0)
        self.assertAlmostEqual(np.sum(weights), 0)

    def test_dendrogram_plot(self):
        """
        Test if dendrogram plot object is correctly rendered.
        """

        hrp = HierarchicalRiskParity()
        hrp.allocate(asset_prices=self.data, asset_names=self.data.columns)
        dendrogram = hrp.plot_clusters(assets=self.data.columns)
        assert dendrogram.get('icoord')
        assert dendrogram.get('dcoord')
        assert dendrogram.get('ivl')
        assert dendrogram.get('leaves')
        assert dendrogram.get('color_list')

    def test_quasi_diagnalization(self):
        """
        Test the quasi-diagnalisation step of HRP algorithm.
        """

        hrp = HierarchicalRiskParity()
        hrp.allocate(asset_prices=self.data, asset_names=self.data.columns)
        assert hrp.ordered_indices == [13, 9, 10, 8, 14, 7, 1, 6, 4, 16, 3, 17,
                                       12, 18, 22, 0, 15, 21, 11, 2, 20, 5, 19]

    def test_value_error_for_non_dataframe_input(self):
        """
        Test ValueError on passing non-dataframe input.
        """

        with self.assertRaises(ValueError):
            hrp = HierarchicalRiskParity()
            hrp.allocate(asset_prices=self.data.values, asset_names=self.data.columns)

    def test_value_error_for_non_date_index(self):
        """
        Test ValueError on passing dataframe not indexed by date.
        """

        with self.assertRaises(ValueError):
            hrp = HierarchicalRiskParity()
            data = self.data.reset_index()
            hrp.allocate(asset_prices=data, asset_names=self.data.columns)

    def test_all_inputs_none(self):
        """
        Test allocation when all inputs are None.
        """

        with self.assertRaises(ValueError):
            hrp = HierarchicalRiskParity()
            hrp.allocate(asset_names=self.data.columns)

    def test_hrp_with_input_as_returns(self):
        """
        Test HRP when passing asset returns dataframe as input.
        """

        hrp = HierarchicalRiskParity()
        returns = ReturnsEstimators().calculate_returns(asset_prices=self.data)
        hrp.allocate(asset_returns=returns, asset_names=self.data.columns)
        weights = hrp.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_hrp_with_input_as_covariance_matrix(self):
        """
        Test HRP when passing a covariance matrix as input.
        """

        hrp = HierarchicalRiskParity()
        returns = ReturnsEstimators().calculate_returns(asset_prices=self.data)
        hrp.allocate(asset_names=self.data.columns, covariance_matrix=returns.cov())
        weights = hrp.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_hrp_with_input_as_distance_matrix(self):
        """
        Test HRP when passing a distance matrix as input.
        """

        hrp = HierarchicalRiskParity()
        returns = ReturnsEstimators().calculate_returns(asset_prices=self.data)
        covariance = returns.cov()
        d_matrix = np.zeros_like(covariance)
        diagnoal_sqrt = np.sqrt(np.diag(covariance))
        np.fill_diagonal(d_matrix, diagnoal_sqrt)
        d_inv = np.linalg.inv(d_matrix)
        corr = np.dot(np.dot(d_inv, covariance), d_inv)
        corr = pd.DataFrame(corr, index=covariance.columns, columns=covariance.columns)
        distance_matrix = np.sqrt((1 - corr).round(5) / 2)
        hrp.allocate(asset_names=self.data.columns, covariance_matrix=covariance, distance_matrix=distance_matrix)
        weights = hrp.weights.values[0]
        self.assertTrue((weights >= 0).all())
        self.assertTrue(len(weights) == self.data.shape[1])
        self.assertAlmostEqual(np.sum(weights), 1)

    def test_hrp_with_linkage_method(self):
        """
        Test HRP when passing a custom linkage method.
        """

        hrp = HierarchicalRiskParity()
        hrp.allocate(asset_names=self.data.columns, asset_prices=self.data, linkage='ward')
        weights = hrp.weights.values[0]
        assert hrp.ordered_indices == [13, 7, 1, 6, 4, 16, 3, 17, 14, 0, 15, 8,
                                       9, 10, 12, 18, 22, 5, 19, 2, 20, 11, 21]
        self.assertTrue((weights >= 0).all())
        self.assertTrue(len(weights) == self.data.shape[1])
        self.assertAlmostEqual(np.sum(weights), 1)

    def test_no_asset_names(self):
        """
        Test HRP when not supplying a list of asset names.
        """

        hrp = HierarchicalRiskParity()
        hrp.allocate(asset_prices=self.data)
        weights = hrp.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_no_asset_names_with_asset_returns(self):
        """
        Test HRP when not supplying a list of asset names and when the user passes asset_returns.
        """

        hrp = HierarchicalRiskParity()
        returns = ReturnsEstimators().calculate_returns(asset_prices=self.data)
        hrp.allocate(asset_returns=returns)
        weights = hrp.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_valuerror_with_no_asset_names(self):
        """
        Test ValueError when not supplying a list of asset names and no other input.
        """

        with self.assertRaises(ValueError):
            hrp = HierarchicalRiskParity()
            returns = ReturnsEstimators().calculate_returns(asset_prices=self.data)
            hrp.allocate(asset_returns=returns.values)



"""
Test various filters.
"""

import unittest
import os
import numpy as np
import pandas as pd

from famlafl.filters.filters import cusum_filter, z_score_filter


class TestCUSUMFilter(unittest.TestCase):
    """
    Tests regarding the CUSUM filter.
    """

    def setUp(self):
        """
        Set the file path for the sample dollar bars data.
        """
        project_path = os.path.dirname(__file__)
        self.path = project_path + '/test_data/dollar_bar_sample.csv'
        self.data = pd.read_csv(self.path, index_col='date_time')
        self.data.index = pd.to_datetime(self.data.index)

    def test_cusum_filter(self):
        """
        Assert that the CUSUM filter works as expected.
        Checks that all the events generated by different threshold values meet the requirements of the filter.
        """

        # Check all events for various threshold levels
        for threshold in [0.005, 0.007, 0.01, 0.015, 0.02, 0.03, 0.04]:
            for timestamp in [True, False]:

                cusum_events = cusum_filter(self.data['close'], threshold=threshold, time_stamps=timestamp)

                for i in range(1, len(cusum_events)):
                    event_1 = self.data.index.get_loc(cusum_events[i - 1])
                    event_2 = self.data.index.get_loc(cusum_events[i])

                    date_range = self.data.iloc[event_1:event_2 + 1]['close']
                    last = np.log(date_range[-1])
                    minimum = np.log(date_range.min())
                    maximum = np.log(date_range.max())

                    # Calculate CUSUM
                    spos = last - minimum
                    sneg = last - maximum
                    cusum = max(np.abs(sneg), spos)

                    self.assertTrue(cusum >= threshold)

    def test_dynamic_cusum_filter(self):
        """
        Test CUSUM filter with dynamic threshold, assert lenght of filtered series
        """
        dynamic_threshold = self.data['close'] * 1e-5
        cusum_events = cusum_filter(self.data['close'], threshold=dynamic_threshold, time_stamps=True)
        self.assertTrue(cusum_events.shape[0] == 9)

    def test_z_score_filter(self):
        """
        Test Z-score filter
        """
        z_score_events = z_score_filter(self.data['close'], 100, 100, 2, time_stamps=True)
        z_score_events_timestamp_false = z_score_filter(self.data['close'], 100, 100, 2, time_stamps=False)

        self.assertTrue(z_score_events.shape[0] == 68)
        self.assertTrue(z_score_events.shape[0] == z_score_events_timestamp_false.shape[0])
        self.assertEqual(self.data.loc[z_score_events[0], 'close'], 2037.25)
        self.assertEqual(self.data.loc[z_score_events[25], 'close'], 2009.5)

    def test_error_raise(self):
        """
        Test ValueError raise of threshold is neither float/int nor pd.Series
        :return:
        """
        with self.assertRaises(ValueError):
            cusum_filter(self.data['close'], threshold='str', time_stamps=True)



# pylint: disable=missing-module-docstring
import numpy as np
from famlafl.online_portfolio_selection.up import UP
from famlafl.online_portfolio_selection.fcorn import FCORN


class FCORNK(UP):
    """
    This class implements the Functional Correlation Driven Nonparametric Learning - K strategy. It
    is reproduced with modification from the following paper:
    `Wang, Y., & Wang, D. (2019). Market Symmetry and Its Application to Pattern-Matching-Based
    Portfolio Selection. The Journal of Financial Data Science, 1(2), 78–92.
    <https://jfds.pm-research.com/content/1/2/78>`_

    Functional Correlation Driven Nonparametric Learning - K formulates a number of FCORN experts and
    tracks the experts performance over time. Each period, the strategy decides to allocate capital
    to the top-k experts until the previous time period. This strategy takes an ensemble approach to
    the top performing experts.
    """

    def __init__(self, window, rho, lambd, k):
        """
        Initializes Functional Correlation Driven Nonparametric Learning - K with the given number
        of window, rho, lambda, and k experts.

        :param window: (int) Number of windows to look back for similarity sets. Generates
                             experts with range of [1, 2, ..., w]. The window ranges typically work
                             well with shorter terms of [1, 7].
        :param rho: (int) Number of rho values for threshold. Generates experts with range of
                          [0, 1, (rho-1)/rho]. Higher rho values allow for a greater coverage of
                          possible parameters, but it will slow down the calculations. Rho ranges
                          typically work well with [3, 5].
        :param lambd: (int) Number of scale factors for sigmoid function. Generates experts with range
                            of [1, 10, 10 ** (lambd - 1)]. Higher lambd values allow for a greater
                            coverage of possible parameters, but it will slow down the calculations.
                            Lambd ranges typically work well with [1, 3].
        :param k: (int) Number of top-k experts. K values have range of [1, window * rho]. Higher
                        number of experts gives a higher exposure to a number of strategies. K value of
                        1 or 2 had the best results as some parameters significantly outperform others.
        """
        self.window = window
        self.rho = rho
        self.lambd = lambd
        self.k = k
        self.number_of_experts = self.window * self.rho * self.lambd
        super().__init__(number_of_experts=self.number_of_experts, weighted='top-k', k=self.k)

    def _initialize(self, asset_prices, weights, resample_by):
        """
        Initializes the important variables for the object.

        :param asset_prices: (pd.DataFrame) Historical asset prices.
        :param weights: (list/np.array/pd.Dataframe) Initial weights set by the user.
        :param resample_by: (str) Specifies how to resample the prices. 'D' for Day, 'W' for Week,
                                  'M' for Month. The inputs are based on pandas' resample method.
        """
        # Check that window value is an integer.
        if not isinstance(self.window, int):
            raise ValueError("Window value must be an integer.")

        # Check that rho value is an integer.
        if not isinstance(self.rho, int):
            raise ValueError("Rho value must be an integer.")

        # Check that lambd value is an integer.
        if not isinstance(self.lambd, int):
            raise ValueError("Lambd value must be an integer.")

        # Check that k value is an integer.
        if not isinstance(self.k, int):
            raise ValueError("K value must be an integer.")

        # Check that window value is at least 1.
        if self.window < 1:
            raise ValueError("Window value must be greater than or equal to 1.")

        # Check that rho value is at least 1.
        if self.rho < 1:
            raise ValueError("Rho value must be greater than or equal to 1.")

        # Check that lambd value is at least 1.
        if self.lambd < 1:
            raise ValueError("Lambd value must be greater than or equal to 0.")

        # Check that k value is at least the number of experts.
        if self.k > self.number_of_experts:
            raise ValueError("K must be less than window * rho * lambd.")

        super(FCORNK, self)._initialize(asset_prices,
                                        weights,
                                        resample_by)

    def _generate_experts(self):
        """
        Generates window * rho experts from window of [1, w], rho of [0, (rho - 1) / rho], and
        lambd of [1, 10 ** (lambd-1)].
        """
        # Initialize expert parameters.
        self.expert_params = np.zeros((self.number_of_experts, 3))
        # Pointer to iterate through parameter locations.
        pointer = 0
        # Window from 1 to self.window.
        for n_window in range(self.window):
            # Rho from 0 to (rho - 1)/rho.
            for n_rho in range(self.rho):
                for n_lambd in range(self.lambd):
                    # Assign experts with parameters (n_window + 1, n_rho/rho, 10 ** n_lambd).
                    self.expert_params[pointer] = [n_window + 1, n_rho / self.rho, 10 ** n_lambd]
                    pointer += 1

        for exp in range(self.number_of_experts):
            param = self.expert_params[exp]
            self.experts.append(
                FCORN(int(param[0]), param[1], param[2]))



"""
Implements the book chapter 7 on Cross Validation for financial data.
"""

from typing import Callable
import pandas as pd
import numpy as np

from sklearn.metrics import log_loss
from sklearn.model_selection import KFold
from sklearn.base import ClassifierMixin
from sklearn.model_selection import BaseCrossValidator


def ml_get_train_times(samples_info_sets: pd.Series, test_times: pd.Series) -> pd.Series:
    # pylint: disable=invalid-name
    """
    Advances in Financial Machine Learning, Snippet 7.1, page 106.

    Purging observations in the training set

    This function find the training set indexes given the information on which each record is based
    and the range for the test set.
    Given test_times, find the times of the training observations.

    :param samples_info_sets: (pd.Series) The information range on which each record is constructed from
        *samples_info_sets.index*: Time when the information extraction started.
        *samples_info_sets.value*: Time when the information extraction ended.
    :param test_times: (pd.Series) Times for the test dataset.
    :return: (pd.Series) Training set
    """
    train = samples_info_sets.copy(deep=True)
    for start_ix, end_ix in test_times.items():
        df0 = train[(start_ix <= train.index) & (train.index <= end_ix)].index  # Train starts within test
        df1 = train[(start_ix <= train) & (train <= end_ix)].index  # Train ends within test
        df2 = train[(train.index <= start_ix) & (end_ix <= train)].index  # Train envelops test
        train = train.drop(df0.union(df1).union(df2))
    return train


class PurgedKFold(KFold):
    """
    Extend KFold class to work with labels that span intervals

    The train is purged of observations overlapping test-label intervals
    Test set is assumed contiguous (shuffle=False), w/o training samples in between

    :param n_splits: (int) The number of splits. Default to 3
    :param samples_info_sets: (pd.Series) The information range on which each record is constructed from
        *samples_info_sets.index*: Time when the information extraction started.
        *samples_info_sets.value*: Time when the information extraction ended.
    :param pct_embargo: (float) Percent that determines the embargo size.
    """

    def __init__(self,
                 n_splits: int = 3,
                 samples_info_sets: pd.Series = None,
                 pct_embargo: float = 0.):

        if not isinstance(samples_info_sets, pd.Series):
            raise ValueError('The samples_info_sets param must be a pd.Series')
        super(PurgedKFold, self).__init__(n_splits, shuffle=False, random_state=None)

        self.samples_info_sets = samples_info_sets
        self.pct_embargo = pct_embargo

    # noinspection PyPep8Naming
    def split(self,
              X: pd.DataFrame,
              y: pd.Series = None,
              groups=None):
        """
        The main method to call for the PurgedKFold class

        :param X: (pd.DataFrame) Samples dataset that is to be split
        :param y: (pd.Series) Sample labels series
        :param groups: (array-like), with shape (n_samples,), optional
            Group labels for the samples used while splitting the dataset into
            train/test set.
        :return: (tuple) [train list of sample indices, and test list of sample indices]
        """
        if X.shape[0] != self.samples_info_sets.shape[0]:
            raise ValueError("X and the 'samples_info_sets' series param must be the same length")

        indices: np.ndarray = np.arange(X.shape[0])
        embargo: int = int(X.shape[0] * self.pct_embargo)

        test_ranges: list[tuple[int, int]] = [(ix[0], ix[-1] + 1) for ix in np.array_split(np.arange(X.shape[0]), self.n_splits)]
        for start_ix, end_ix in test_ranges:
            test_indices = indices[start_ix:end_ix]

            if end_ix < X.shape[0]:
                end_ix += embargo

            test_times = pd.Series(index=[self.samples_info_sets[start_ix]], data=[self.samples_info_sets[end_ix-1]])
            train_times = ml_get_train_times(self.samples_info_sets, test_times)

            train_indices = []
            for train_ix in train_times.index:
                train_indices.append(self.samples_info_sets.index.get_loc(train_ix))
            yield np.array(train_indices), test_indices


# noinspection PyPep8Naming
def ml_cross_val_score(
        classifier: ClassifierMixin,
        X: pd.DataFrame,
        y: pd.Series,
        cv_gen: BaseCrossValidator,
        sample_weight_train: np.ndarray = None,
        sample_weight_score: np.ndarray = None,
        scoring: Callable[[np.array, np.array], float] = log_loss):
    # pylint: disable=invalid-name
    # pylint: disable=comparison-with-callable
    """
    Advances in Financial Machine Learning, Snippet 7.4, page 110.

    Using the PurgedKFold Class.

    Function to run a cross-validation evaluation of the using sample weights and a custom CV generator.

    Note: This function is different to the book in that it requires the user to pass through a CV object. The book
    will accept a None value as a default and then resort to using PurgedCV, this also meant that extra arguments had to
    be passed to the function. To correct this we have removed the default and require the user to pass a CV object to
    the function.

    Example:

    .. code-block:: python

        cv_gen = PurgedKFold(n_splits=n_splits, samples_info_sets=samples_info_sets, pct_embargo=pct_embargo)
        scores_array = ml_cross_val_score(classifier, X, y, cv_gen, sample_weight_train=sample_train,
                                          sample_weight_score=sample_score, scoring=accuracy_score)

    :param classifier: (ClassifierMixin) A sk-learn Classifier object instance.
    :param X: (pd.DataFrame) The dataset of records to evaluate.
    :param y: (pd.Series) The labels corresponding to the X dataset.
    :param cv_gen: (BaseCrossValidator) Cross Validation generator object instance.
    :param sample_weight_train: (np.array) Sample weights used to train the model for each record in the dataset.
    :param sample_weight_score: (np.array) Sample weights used to evaluate the model quality.
    :param scoring: (Callable) A metric scoring, can be custom sklearn metric.
    :return: (np.array) The computed score.
    """

    # If no sample_weight then broadcast a value of 1 to all samples (full weight).
    if sample_weight_train is None:
        sample_weight_train = np.ones((X.shape[0],))

    if sample_weight_score is None:
        sample_weight_score = np.ones((X.shape[0],))

    # Score model on KFolds
    ret_scores = []
    for train, test in cv_gen.split(X=X, y=y):
        fit = classifier.fit(X=X.iloc[train, :], y=y.iloc[train], sample_weight=sample_weight_train[train])
        if scoring == log_loss:
            prob = fit.predict_proba(X.iloc[test, :])
            # Negate log_loss since sklearn metrics are optimized for higher values
            score = -scoring(y.iloc[test], prob, sample_weight=sample_weight_score[test], labels=classifier.classes_)
        else:
            pred = fit.predict(X.iloc[test, :])
            score = scoring(y.iloc[test], pred, sample_weight=sample_weight_score[test])
        ret_scores.append(score)
    return np.array(ret_scores)



"""
This module contains functionality for determining bet sizes for investments based on machine learning predictions.
These implementations are based on bet sizing approaches described in Chapter 10.
"""

import numpy as np
import pandas as pd
from scipy.stats import norm, moment

from famlafl.bet_sizing.ch10_snippets import get_signal, avg_active_signals, discrete_signal
from famlafl.bet_sizing.ch10_snippets import get_w, get_target_pos, limit_price, bet_size
from famlafl.bet_sizing.ef3m import M2N, raw_moment, most_likely_parameters


def bet_size_probability(events, prob, num_classes, pred=None, step_size=0.0, average_active=False, num_threads=1):
    """
    Calculates the bet size using the predicted probability. Note that if 'average_active' is True, the returned
    pandas.Series will be twice the length of the original since the average is calculated at each bet's open and close.

    :param events: (pandas.DataFrame) Contains at least the column 't1', the expiry datetime of the product, with
     a datetime index, the datetime the position was taken.
    :param prob: (pandas.Series) The predicted probability.
    :param num_classes: (int) The number of predicted bet sides.
    :param pred: (pd.Series) The predicted bet side. Default value is None which will return a relative bet size
     (i.e. without multiplying by the side).
    :param step_size: (float) The step size at which the bet size is discretized, default is 0.0 which imposes no
     discretization.
    :param average_active: (bool) Option to average the size of active bets, default value is False.
    :param num_threads: (int) The number of processing threads to utilize for multiprocessing, default value is 1.
    :return: (pandas.Series) The bet size, with the time index.
    """
    signal_0 = get_signal(prob, num_classes, pred)
    events_0 = signal_0.to_frame('signal').join(events['t1'], how='left')
    if average_active:
        signal_1 = avg_active_signals(events_0, num_threads)
    else:
        signal_1 = events_0.signal

    if abs(step_size) > 0:
        signal_1 = discrete_signal(signal0=signal_1, step_size=abs(step_size))

    return signal_1


def bet_size_dynamic(current_pos, max_pos, market_price, forecast_price, cal_divergence=10, cal_bet_size=0.95,
                     func='sigmoid'):
    """
    Calculates the bet sizes, target position, and limit price as the market price and forecast price fluctuate.
    The current position, maximum position, market price, and forecast price can be passed as separate pandas.Series
    (with a common index), as individual numbers, or a combination thereof. If any one of the aforementioned arguments
    is a pandas.Series, the other arguments will be broadcast to a pandas.Series of the same length and index.

    :param current_pos: (pandas.Series, int) Current position.
    :param max_pos: (pandas.Series, int) Maximum position
    :param market_price: (pandas.Series, float) Market price.
    :param forecast_price: (pandas.Series, float) Forecast price.
    :param cal_divergence: (float) The divergence to use in calibration.
    :param cal_bet_size: (float) The bet size to use in calibration.
    :param func: (string) Function to use for dynamic calculation. Valid options are: 'sigmoid', 'power'.
    :return: (pandas.DataFrame) Bet size (bet_size), target position (t_pos), and limit price (l_p).
    """
    # Create a dictionary of bet size variables for easier handling.
    d_vars = {'pos': current_pos, 'max_pos': max_pos, 'm_p': market_price, 'f': forecast_price}
    events_0 = confirm_and_cast_to_df(d_vars)

    # Calibrate w.
    w_param = get_w(cal_divergence, cal_bet_size, func)
    # Compute the target bet position.
    events_0['t_pos'] = events_0.apply(lambda x: get_target_pos(w_param, x.f, x.m_p, x.max_pos, func), axis=1)
    # Compute the break even limit price.
    events_0['l_p'] = events_0.apply(lambda x: limit_price(x.t_pos, x.pos, x.f, w_param, x.max_pos, func), axis=1)
    # Compute the bet size.
    events_0['bet_size'] = events_0.apply(lambda x: bet_size(w_param, x.f-x.m_p, func), axis=1)

    return events_0[['bet_size', 't_pos', 'l_p']]


def bet_size_budget(events_t1, sides):
    """
    Calculates a bet size from the bet sides and start and end times. These sequences are used to determine the
    number of concurrent long and short bets, and the resulting strategy-independent bet sizes are the difference
    between the average long and short bets at any given time. This strategy is based on the section 10.2
    in "Advances in Financial Machine Learning". This creates a linear bet sizing scheme that is aligned to the
    expected number of concurrent bets in the dataset.

    :param events_t1: (pandas.Series) The end datetime of the position with the start datetime as the index.
    :param sides: (pandas.Series) The side of the bet with the start datetime as index. Index must match the
     'events_t1' argument exactly. Bet sides less than zero are interpretted as short, bet sides greater than zero
     are interpretted as long.
    :return: (pandas.DataFrame) The 'events_t1' and 'sides' arguments as columns, with the number of concurrent
     active long and short bets, as well as the bet size, in additional columns.
    """
    events_1 = get_concurrent_sides(events_t1, sides)
    active_long_max, active_short_max = events_1['active_long'].max(), events_1['active_short'].max()
    frac_active_long = events_1['active_long'] / active_long_max if active_long_max > 0 else 0
    frac_active_short = events_1['active_short'] / active_short_max if active_short_max > 0 else 0
    events_1['bet_size'] = frac_active_long - frac_active_short

    return events_1


def bet_size_reserve(events_t1, sides, fit_runs=100, epsilon=1e-5, factor=5, variant=2, max_iter=10_000,
                     num_workers=1, return_parameters=False):
    """
    Calculates the bet size from bet sides and start and end times. These sequences are used to determine the number
    of concurrent long and short bets, and the difference between the two at each time step, c_t. A mixture of two
    Gaussian distributions is fit to the distribution of c_t, which is then used to determine the bet size. This
    strategy results in a sigmoid-shaped bet sizing response aligned to the expected number of concurrent long
    and short bets in the dataset.

    Note that this function creates a <famlafl.bet_sizing.ef3m.M2N> object and makes use of the parallel fitting
    functionality. As such, this function accepts and passes fitting parameters to the
    famlafl.bet_sizing.ef3m.M2N.mp_fit() method.

    :param events_t1: (pandas.Series) The end datetime of the position with the start datetime as the index.
    :param sides: (pandas.Series) The side of the bet with the start datetime as index. Index must match the
     'events_t1' argument exactly. Bet sides less than zero are interpretted as short, bet sides greater than zero
     are interpretted as long.
    :param fit_runs: (int) Number of runs to execute when trying to fit the distribution.
    :param epsilon: (float) Error tolerance.
    :param factor: (float) Lambda factor from equations.
    :param variant: (int) Which algorithm variant to use, 1 or 2.
    :param max_iter: (int) Maximum number of iterations after which to terminate loop.
    :param num_workers: (int) Number of CPU cores to use for multiprocessing execution, set to -1 to use all
     CPU cores. Default is 1.
    :param return_parameters: (bool) If True, function also returns a dictionary of the fited mixture parameters.
    :return: (pandas.DataFrame) The 'events_t1' and 'sides' arguments as columns, with the number of concurrent
     active long, short bets, the difference between long and short, and the bet size in additional columns.
     Also returns the mixture parameters if 'return_parameters' is set to True.
    """
    events_active = get_concurrent_sides(events_t1, sides)
    # Calculate the concurrent difference in active bets: c_t = <current active long> - <current active short>
    events_active['c_t'] = events_active['active_long'] - events_active['active_short']
    # Calculate the first 5 centered and raw moments from the c_t distribution.
    central_mmnts = [moment(events_active['c_t'].to_numpy(), moment=i) for i in range(1, 6)]
    raw_mmnts = raw_moment(central_moments=central_mmnts, dist_mean=events_active['c_t'].mean())
    # Fit the mixture of distributions.
    m2n = M2N(raw_mmnts, epsilon=epsilon, factor=factor, n_runs=fit_runs,
              variant=variant, max_iter=max_iter, num_workers=num_workers)
    df_fit_results = m2n.mp_fit()
    fit_params = most_likely_parameters(df_fit_results)
    params_list = [fit_params[key] for key in ['mu_1', 'mu_2', 'sigma_1', 'sigma_2', 'p_1']]
    # Calculate the bet size.
    events_active['bet_size'] = events_active['c_t'].apply(lambda c: single_bet_size_mixed(c, params_list))

    if return_parameters:
        return events_active, fit_params
    return events_active


def confirm_and_cast_to_df(d_vars):
    """
    Accepts either pandas.Series (with a common index) or integer/float values, casts all non-pandas.Series values
    to Series, and returns a pandas.DataFrame for further calculations. This is a helper function to the
    'bet_size_dynamic' function.

    :param d_vars: (dict) A dictionary where the values are either pandas.Series or single int/float values.
     All pandas.Series passed are assumed to have the same index. The keys of the dictionary will be used for column
     names in the returned pandas.DataFrame.
    :return: (pandas.DataFrame) The values from the input dictionary in pandas.DataFrame format, with dictionary
     keys as column names.
    """
    any_series = False  # Are any variables a pandas.Series?
    all_series = True  # Are all variables a pandas.Series?
    ser_len = 0
    for var in d_vars.values():
        any_series = any_series or isinstance(var, pd.Series)
        all_series = all_series and isinstance(var, pd.Series)

        if isinstance(var, pd.Series):
            ser_len = var.size
            idx = var.index

    # Handle data types if there are no pandas.Series variables.
    if not any_series:
        for k in d_vars:
            d_vars[k] = pd.Series(data=[d_vars[k]], index=[0])

    # Handle data types if some but not all variables are pandas.Series.
    if any_series and not all_series:
        for k in d_vars:
            if not isinstance(d_vars[k], pd.Series):
                d_vars[k] = pd.Series(data=np.array([d_vars[k] for i in range(ser_len)]), index=idx)

    # Combine Series to form a DataFrame.
    events = pd.concat(list(d_vars.values()), axis=1)
    events.columns = list(d_vars.keys())

    return events


def get_concurrent_sides(events_t1, sides):
    """
    Given the side of the position along with its start and end timestamps, this function returns two pandas.Series
    indicating the number of concurrent long and short bets at each timestamp.

    :param events_t1: (pandas.Series) The end datetime of the position with the start datetime as the index.
    :param sides: (pandas.Series) The side of the bet with the start datetime as index. Index must match the
     'events_t1' argument exactly. Bet sides less than zero are interpretted as short, bet sides greater than zero
     are interpretted as long.
    :return: (pandas.DataFrame) The 'events_t1' and 'sides' arguments as columns, with two additional columns
     indicating the number of concurrent active long and active short bets at each timestamp.
    """
    events_0 = pd.DataFrame({'t1':events_t1, 'side':sides})
    events_0['active_long'] = 0
    events_0['active_short'] = 0

    for idx in events_0.index:
        # A bet side greater than zero indicates a long position.
        df_long_active_idx = set(events_0[(events_0.index <= idx) & (events_0['t1'] > idx) & (events_0['side'] > 0)].index)
        events_0.loc[idx, 'active_long'] = len(df_long_active_idx)
        # A bet side less than zero indicates a short position.
        df_short_active_idx = set(events_0[(events_0.index <= idx) & (events_0['t1'] > idx) & (events_0['side'] < 0)].index)
        events_0.loc[idx, 'active_short'] = len(df_short_active_idx)

    return events_0


def cdf_mixture(x_val, parameters):
    """
    The cumulative distribution function of a mixture of 2 normal distributions, evaluated at x_val.

    :param x_val: (float) Value at which to evaluate the CDF.
    :param parameters: (list) The parameters of the mixture, [mu_1, mu_2, sigma_1, sigma_2, p_1]
    :return: (float) CDF of the mixture.
    """
    mu_1, mu_2, sigma_1, sigma_2, p_1 = parameters  # Parameters reassigned for clarity.
    return p_1 * norm.cdf(x_val, mu_1, sigma_1) + (1-p_1) * norm.cdf(x_val, mu_2, sigma_2)


def single_bet_size_mixed(c_t, parameters):
    """
    Returns the single bet size based on the description provided in question 10.4(c), provided the difference in
    concurrent long and short positions, c_t, and the fitted parameters of the mixture of two Gaussain distributions.

    :param c_t: (int) The difference in the number of concurrent long bets minus short bets.
    :param parameters: (list) The parameters of the mixture, [mu_1, mu_2, sigma_1, sigma_2, p_1]
    :return: (float) Bet size.
    """
    if c_t >= 0:
        single_bet_size = (cdf_mixture(c_t, parameters) - cdf_mixture(0, parameters)) / (1 - cdf_mixture(0, parameters))
    else:
        single_bet_size = (cdf_mixture(c_t, parameters) - cdf_mixture(0, parameters)) / cdf_mixture(0, parameters)
    return single_bet_size



"""
Module which implements feature importance algorithms described in Chapter 8 and other interpretability tools
from the Journal of Financial Data Science.
"""

from famlafl.feature_importance.importance import (mean_decrease_impurity, mean_decrease_accuracy,
                                                    single_feature_importance, plot_feature_importance)
from famlafl.feature_importance.orthogonal import (feature_pca_analysis, get_pca_rank_weighted_kendall_tau,
                                                    get_orthogonal_features)
from famlafl.feature_importance.fingerpint import RegressionModelFingerprint, ClassificationModelFingerprint



"""
Module implementing typical financial datasets load (stock prices, dollar bars, ticks)
"""
from famlafl.datasets.load_datasets import (load_dollar_bar_sample, load_stock_prices, load_tick_sample)



"""
Functions derived from Chapter 10: Bet Sizing
Only the highest-level user functions are included in the __init__ file.
"""

from famlafl.bet_sizing.bet_sizing import (bet_size_probability, bet_size_dynamic, bet_size_budget, bet_size_reserve,
                                            confirm_and_cast_to_df, get_concurrent_sides, cdf_mixture,
                                            single_bet_size_mixed)
from famlafl.bet_sizing.ef3m import M2N, centered_moment, raw_moment, most_likely_parameters



"""
Test various volatility estimates
"""

import unittest
import os
import pandas as pd

from famlafl.util import get_parksinson_vol, get_yang_zhang_vol, get_garman_class_vol


class TestVolatilityEstimators(unittest.TestCase):
    """
    Test various volatility estimates (YZ, GS, Parksinson)
    """

    def setUp(self):
        """
        Set the file path for the sample dollar bars data.
        """
        project_path = os.path.dirname(__file__)
        self.path = project_path + '/test_data/dollar_bar_sample.csv'
        self.trades_path = project_path + '/test_data/tick_data.csv'
        self.data = pd.read_csv(self.path, index_col='date_time', parse_dates=[0])
        self.data.index = pd.to_datetime(self.data.index)

    def test_vol_features(self):
        """
        Test volatility estimators
        """
        gm_vol = get_garman_class_vol(self.data.open, self.data.high, self.data.low, self.data.close, window=20)
        yz_vol = get_yang_zhang_vol(self.data.open, self.data.high, self.data.low, self.data.close, window=20)
        park_vol = get_parksinson_vol(self.data.high, self.data.low, window=20)

        self.assertEqual(self.data.shape[0], gm_vol.shape[0])
        self.assertEqual(self.data.shape[0], yz_vol.shape[0])
        self.assertEqual(self.data.shape[0], park_vol.shape[0])

        self.assertAlmostEqual(gm_vol.mean(), 0.001482, delta=1e-6)
        self.assertAlmostEqual(yz_vol.mean(), 0.00162001, delta=1e-6)
        self.assertAlmostEqual(park_vol.mean(), 0.00149997, delta=1e-6)



# famlafl/util/multiprocess.py
"""
Contains functionality for multiprocessing.
"""

import sys
import datetime as dt
import time
from multiprocessing import Process, Queue, cpu_count
import numpy as np
import pandas as pd


def expand_call(kwargs):
    """
    Expand the arguments of a callback function, python.apply_async allows only
    one argument, tuple.
    """
    func = kwargs['func']
    del kwargs['func']
    if 'queue' in kwargs:
        queue = kwargs.pop('queue')
        try:
            out = func(**kwargs)
            if queue is not None:
                queue.put(('success', out))
        except Exception as e:
            if queue is not None:
                queue.put(('error', str(e)))
        return None
    return func(**kwargs)


def report_progress(job_number, num_jobs, time0, task):
    """
    Report progress as asynch jobs are completed.

    :param job_number: (int) Number of jobs completed
    :param num_jobs: (int) Total number of jobs
    :param time0: (time) Start time
    :param task: (str) Task description
    """
    # 1) Compute progress
    msg = [float(job_number) / num_jobs, (time.time() - time0) / 60.0]
    # 2) Add message
    msg.append(msg[1] * (1 / msg[0] - 1))
    time_stamp = str(dt.datetime.fromtimestamp(time.time()))

    # 3) Report progress
    if job_number < num_jobs:
        sys.stderr.write(
            time_stamp + ' ' + task + ' ' +
            str(round(msg[0] * 100, 2)) + '% done after ' + str(
                round(msg[1], 2)) + ' minutes. Remaining ' + str(
                    round(msg[2], 2)) + ' minutes.')
    else:
        sys.stderr.write(
            time_stamp + ' ' + task + ' done after ' + str(round(msg[1], 2)) +
            ' minutes.')


def process_jobs(jobs, task=None, num_threads=cpu_count()):
    """
    Run in parallel. jobs must contain a 'func' callback, for expand_call

    :param jobs: (list) Jobs (each job is a dict)
    :param task: (str) Task description
    :param num_threads: (int) The number of threads that will be used in parallel (one processor per thread)
    :return: (pd.Series or pd.DataFrame) Returns a pandas object with the results
    """
    if not jobs:
        # Return empty Series for empty jobs list
        return pd.Series(dtype=float)
        
    if num_threads == 1:
        # Run jobs sequentially for better error handling
        outputs = []
        for job in jobs:
            try:
                result = expand_call(job)
                if result is not None:
                    outputs.append(result)
            except Exception as e:
                print(f"Error in job: {str(e)}")
                # Return empty DataFrame with default columns for error handling
                return pd.DataFrame(columns=['t1', 'pt', 'sl'])
    else:
        if task is None:
            task = jobs[0]['func'].__name__
        queue = Queue()
        processes = []
        for job in jobs:
            job_copy = job.copy()
            job_copy['queue'] = queue
            process = Process(target=expand_call, args=(job_copy,))
            processes.append(process)

        # Additional variables for controlling the number of concurrent processes
        processes_alive = 0  # Counter of alive processes
        process_idx = 0  # Index of the next process to start
        outputs = []  # Stores outputs as they arrive
        time0 = time.time()

        # Start processes
        while True:
            # Start processes
            while processes_alive < num_threads and process_idx < len(processes):
                processes[process_idx].start()
                process_idx += 1
                processes_alive += 1

            # Get output
            if not queue.empty():
                status, result = queue.get()
                if status == 'success':
                    if result is not None:
                        outputs.append(result)
                else:  # status == 'error'
                    print(f"Error in job: {result}")
                report_progress(len(outputs), len(processes), time0, task)

            # Check if any process is done
            processes_alive = sum([process.is_alive() for process in processes])

            # Exit if all processes are done
            if processes_alive == 0 and process_idx == len(processes):
                break

            # Sleep for a short time before checking again
            time.sleep(0.1)

    # Handle outputs
    if len(outputs) > 0:
        if isinstance(outputs[0], pd.DataFrame):
            # For DataFrames, preserve job structure
            return outputs
        elif isinstance(outputs[0], pd.Series):
            # For Series, concatenate all outputs
            return pd.concat(outputs)
        # For non-pandas outputs, always preserve job structure
        return outputs
    # Return empty DataFrame with default columns for empty outputs
    return pd.DataFrame(columns=['t1', 'pt', 'sl'])


def process_jobs_(jobs, task=None, num_threads=cpu_count()):
    """
    Run in parallel. jobs must contain a 'func' callback, for expand_call.
    This is a variant of process_jobs that returns both outputs and failed jobs.

    :param jobs: (list) Jobs (each job is a dict)
    :param task: (str) Task description
    :param num_threads: (int) The number of threads that will be used in parallel (one processor per thread)
    :return: (tuple) The first element is the outputs list, the second element is the failed jobs list
    """
    if task is None:
        task = jobs[0]['func'].__name__
    queue = Queue()
    processes = []
    for job in jobs:
        job_copy = job.copy()
        job_copy['queue'] = queue
        process = Process(target=expand_call, args=(job_copy,))
        processes.append(process)

    # Additional variables for controlling the number of concurrent processes
    processes_alive = 0  # Counter of alive processes
    process_idx = 0  # Index of the next process to start
    outputs = []  # Stores outputs as they arrive
    failed_indices = set()  # Stores indices of failed jobs
    time0 = time.time()

    # Start processes
    while True:
        # Start processes
        while processes_alive < num_threads and process_idx < len(processes):
            processes[process_idx].start()
            process_idx += 1
            processes_alive += 1

        # Get output
        if not queue.empty():
            try:
                status, result = queue.get()
                job_idx = len(outputs) + len(failed_indices)  # Current job index
                if status == 'success':
                    if result is not None:
                        outputs.append(result)
                    else:
                        failed_indices.add(job_idx)
                else:  # status == 'error'
                    failed_indices.add(job_idx)
                    print(f"Error in job: {result}")
                report_progress(job_idx + 1, len(processes), time0, task)
            except Exception as e:
                # If we can't get the result, mark the current job as failed
                failed_indices.add(len(outputs) + len(failed_indices))
                print(f"Error getting result: {str(e)}")

        # Check if any process is done
        processes_alive = sum([process.is_alive() for process in processes])

        # Exit if all processes are done
        if processes_alive == 0 and process_idx == len(processes):
            # Ensure all remaining jobs are marked as failed
            while len(outputs) + len(failed_indices) < len(jobs):
                failed_indices.add(len(outputs) + len(failed_indices))
            break

        # Sleep for a short time before checking again
        time.sleep(0.1)

    # Convert failed indices to failed jobs list
    failed = [jobs[i] for i in sorted(failed_indices)]

    # Handle outputs
    if len(outputs) > 0:
        if isinstance(outputs[0], pd.DataFrame):
            outputs = pd.concat(outputs, axis=0)
        elif isinstance(outputs[0], pd.Series):
            # For Series, preserve job structure (one Series per successful job)
            outputs = outputs  # Keep all successful outputs
    elif len(jobs) > 0:
        # If no outputs but jobs exist, create empty Series for successful jobs
        outputs = [pd.Series(index=jobs[i]['molecule']) for i in range(len(jobs)) if i not in failed_indices]
    
    return outputs, failed


def process_jobs_in_batches(jobs, task=None, num_threads=cpu_count(), batch_size=1, verbose=False):
    """
    Run in parallel. jobs must contain a 'func' callback, for expand_call. The jobs will be processed in batches.

    :param jobs: (list) Jobs (each job is a dict)
    :param task: (str) Task description
    :param num_threads: (int) The number of threads that will be used in parallel (one processor per thread)
    :param batch_size: (int) Number of jobs to process in each batch
    :param verbose: (bool) Flag to report progress on jobs or not
    """
    results = []  # All results
    failed = []  # Failed jobs

    if task is None:
        task = jobs[0]['func'].__name__

    # Process jobs in batches
    for batch_num, i in enumerate(range(0, len(jobs), batch_size)):
        batch_jobs = jobs[i:i + batch_size]
        if verbose:
            print(f"Processing batch {batch_num + 1}")

        queue = Queue()
        processes = []
        for job in batch_jobs:
            job_copy = job.copy()
            job_copy['queue'] = queue
            process = Process(target=expand_call, args=(job_copy,))
            processes.append(process)

        # Additional variables for controlling the number of concurrent processes
        processes_alive = 0  # Counter of alive processes
        process_idx = 0  # Process index
        batch_outputs = []  # Stores outputs for current batch
        time0 = time.time()

        # Start processes
        while True:
            # Start processes
            while processes_alive < num_threads and process_idx < len(processes):
                processes[process_idx].start()
                process_idx += 1
                processes_alive += 1

            # Get output
            if not queue.empty():
                status, result = queue.get()
                if status == 'success':
                    if result is not None:
                        batch_outputs.append(result)
                    else:
                        failed.append(batch_jobs[len(batch_outputs)])
                else:  # status == 'error'
                    failed.append(batch_jobs[len(batch_outputs)])
                    print(f"Error in job: {result}")

                if verbose:
                    report_progress(len(batch_outputs), len(processes), time0, task)

            # Check if any process is done
            processes_alive = sum([process.is_alive() for process in processes])

            # Exit if all processes are done
            if processes_alive == 0 and process_idx == len(processes):
                break

            # Sleep for a short time before checking again
            time.sleep(0.1)

        # Handle batch outputs
        if len(batch_outputs) > 0:
            if isinstance(batch_outputs[0], pd.DataFrame):
                results.append(pd.concat(batch_outputs, axis=0))
            elif isinstance(batch_outputs[0], pd.Series):
                results.append(pd.concat(batch_outputs))
            else:
                results.extend(batch_outputs)

    # Combine all results
    if len(results) > 0:
        if isinstance(results[0], pd.DataFrame):
            return pd.concat(results, axis=0), failed
        elif isinstance(results[0], pd.Series):
            return pd.concat(results), failed
    return results, failed


def lin_parts(num_atoms, num_threads):
    """
    Partition of atoms with a linear logic.

    :param num_atoms: (int) Number of atoms
    :param num_threads: (int) Number of threads that will be used in parallel (one processor per thread)
    :return: (np.array) Partition of atoms
    """
    # Partition of atoms with a linear logic
    parts = np.linspace(0, num_atoms, min(num_threads, num_atoms) + 1)
    parts = np.ceil(parts).astype(int)
    return parts


def nested_parts(num_atoms, num_threads, upper_triangle=False):
    """
    Partition of atoms with a nested logic.

    :param num_atoms: (int) Number of atoms
    :param num_threads: (int) Number of threads that will be used in parallel (one processor per thread)
    :param upper_triangle: (bool) Flag to partition an upper triangular matrix
    :return: (np.array) Partition of atoms
    """
    # Partition of atoms with a nested logic
    parts = [0]
    num_threads_ = min(num_threads, num_atoms)

    for _ in range(num_threads_):
        part = 1 + 4 * (parts[-1] ** 2 + parts[-1] + num_atoms * (num_atoms + 1.0) / num_threads_)
        part = (-1 + part ** 0.5) / 2.0
        parts.append(part)

    parts = np.round(parts).astype(int)

    if upper_triangle:  # Partition for an upper triangular matrix
        parts = np.cumsum(np.diff(parts)[::-1])
        parts = np.append(np.array([0]), parts)

    return parts


def mp_pandas_obj(func, pd_obj, num_threads=24, mp_batches=1, lin_mols=True, **kargs):
    """
    Parallelize jobs, return a dataframe or series.
    Example: df1=mp_pandas_obj(func,('molecule',df0.index),24,**kwds)

    :param func: (function) A function to be parallelized
    :param pd_obj: (tuple) Element 0: The name of the argument used to pass the molecule
                          Element 1: A pandas object
    :param num_threads: (int) The number of threads that will be used in parallel (one processor per thread)
    :param mp_batches: (int) Number of batches
    :param lin_mols: (bool) Tells if the method should use linear or nested partitioning
    :param kargs: (var args) Keyword arguments to be passed to the function
    :return: (pd.DataFrame) Returns a pandas object with the results
    """
    # Handle empty input case
    if len(pd_obj[1]) == 0:
        return pd.Series(dtype=float)

    # Ensure at least one batch and valid number of threads
    mp_batches = max(1, min(mp_batches, len(pd_obj[1])))
    num_threads = min(num_threads, len(pd_obj[1]))

    # Create parts for the entire dataset
    if lin_mols:
        parts = lin_parts(len(pd_obj[1]), num_threads)
    else:
        parts = nested_parts(len(pd_obj[1]), num_threads)

    # Create jobs based on parts
    jobs = []
    for i in range(1, len(parts)):
        job = {pd_obj[0]: pd_obj[1][parts[i - 1]:parts[i]], 'func': func}
        job.update(kargs)
        jobs.append(job)

    if not jobs:  # If no jobs were created
        # Handle single item case
        if len(pd_obj[1]) == 1:
            job = {pd_obj[0]: pd_obj[1], 'func': func}
            job.update(kargs)
            jobs.append(job)
        else:
            return pd.Series(dtype=float)

    # Process all jobs at once
    result = process_jobs(jobs, num_threads=num_threads)
    
    # Handle different types of results
    if isinstance(result, (pd.Series, pd.DataFrame)):
        return result
    elif isinstance(result, list):
        if all(isinstance(x, pd.DataFrame) for x in result):
            return pd.concat(result, axis=0)
        elif all(isinstance(x, pd.Series) for x in result):
            # For Series results, concatenate and ensure index is preserved
            concatenated = pd.concat(result)
            # If the original index was split across jobs, ensure we get all values
            if len(concatenated) < len(pd_obj[1]):
                return pd.Series(index=pd_obj[1])
            return concatenated
        # For non-pandas outputs, preserve original structure
        return result
    
    return pd.Series(dtype=float)


"""
Tests Buy and Hold.
"""
from unittest import TestCase
import os
import numpy as np
import pandas as pd
from famlafl.online_portfolio_selection import BAH


class TestBuyAndHold(TestCase):
    # pylint: disable=too-many-public-methods
    # pylint: disable=E1136
    """
    Tests different functions of the Buy and Hold class.
    """

    def setUp(self):
        """
        Sets the file path for the tick data csv.
        """
        # Set project path to current directory.
        project_path = os.path.dirname(__file__)
        # Add new data path to match stock_prices.csv data.
        data_path = project_path + '/test_data/stock_prices.csv'
        # Read csv, parse dates, and drop NaN.
        self.data = pd.read_csv(data_path, parse_dates=True, index_col="Date").dropna(axis=1)

    def test_buy_and_hold_solution(self):
        """
        Tests the calculation of buy and hold weights.
        """
        # Initialize BAH.
        bah = BAH()
        # Allocates asset prices to BAH.
        bah.allocate(self.data)
        # Create np.array of all_weights.
        all_weights = np.array(bah.all_weights)
        # Check if all weights sum to 1.
        for i in range(all_weights.shape[0]):
            weights = all_weights[i]
            assert (weights >= 0).all()
            assert len(weights) == self.data.shape[1]
            np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_buy_two_assets(self):
        """
        Tests that weights are changing for a portfolio of two assets.
        """
        # Initialize weights with 0.5 on two random assets.
        weight = np.zeros(self.data.shape[1])
        weight[:2] = 0.5
        # Initialize BAH.
        bah1 = BAH()
        # Allocates asset prices to BAH.
        bah1.allocate(self.data, weight)
        # Get the second to last weight.
        last_bah1_weight = np.array(bah1.all_weights)[-1]
        # Relative price is calculated by dividing the second to last row by the first row.
        price_diff = np.array(self.data.iloc[-2] / self.data.iloc[0])
        # Calculate portfolio growth.
        new_last_weight = price_diff * weight
        # Normalize to sum the weight to one.
        norm_new_weight = new_last_weight/np.sum(new_last_weight)
        # Manual calculation should equal the weights returned by BAH.
        np.testing.assert_almost_equal(last_bah1_weight, norm_new_weight)

    def test_buy_five_assets(self):
        """
        Tests that weights are changing for a portfolio of five assets.
        """
        # Initialize weights with 0.2 on five random assets.
        weight = np.zeros(self.data.shape[1])
        weight[:2] = 0.5
        # Initialize BAH.
        bah2 = BAH()
        # Allocates asset prices to BAH.
        bah2.allocate(self.data, weight)
        # Get the second to last weight.
        last_bah2_weight = np.array(bah2.all_weights)[-1]
        # Relative price is calculated by dividing the second to last row by the first row.
        price_diff = np.array(self.data.iloc[-2] / self.data.iloc[0])
        # Calculate portfolio growth.
        new_last_weight = price_diff * weight
        # Normalize to sum the weight to one.
        norm_new_weight = new_last_weight/np.sum(new_last_weight)
        # Manual calculation should equal the weights returned by BAH.
        np.testing.assert_almost_equal(last_bah2_weight, norm_new_weight)
